"""Some helper function for PyGeoHydro."""

# pyright: reportGeneralTypeIssues=false
from __future__ import annotations

import io
from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, Any, Tuple, Union, cast

import cytoolz.curried as tlz
import geopandas as gpd
import numpy as np
import pandas as pd
import ujson as json
from defusedxml import ElementTree

import async_retriever as ar
from pygeohydro import us_abbrs
from pygeohydro.exceptions import InputRangeError, InputTypeError, InputValueError
from pygeoogc import ServiceURL

if TYPE_CHECKING:
    import pyproj
    from shapely import MultiPolygon, Polygon

    GTYPE = Union[Polygon, MultiPolygon, Tuple[float, float, float, float]]
    CRSTYPE = Union[int, str, pyproj.CRS]
__all__ = ["nlcd_helper", "nwis_errors", "states_lookup_table", "get_us_states"]


def nlcd_helper() -> dict[str, Any]:
    """Get legends and properties of the NLCD cover dataset.

    Notes
    -----
    The following references have been used:
        - https://github.com/jzmiller1/nlcd
        - https://www.mrlc.gov/data-services-page
        - https://www.mrlc.gov/data/legends/national-land-cover-database-2016-nlcd2016-legend
        - https://doi.org/10.1111/jfr3.12347

    Returns
    -------
    dict
        Years when data is available and cover classes and categories, and roughness estimations.
    """
    base_url = "https://www.mrlc.gov/downloads/sciweb1/shared/mrlc/metadata"
    base_path = "eainfo/detailed/attr/attrdomv/edom"

    def _get_xml(
        layer: str,
    ) -> tuple[Any, Any, Any]:
        et = ElementTree.fromstring(ar.retrieve_text([f"{base_url}/{layer}.xml"], ssl=False)[0])
        return (
            et,
            et.findall(f"{base_path}/edomv"),
            et.findall(f"{base_path}/edomvd"),
        )

    root, edomv, edomvd = _get_xml("NLCD_2019_Land_Cover_Science_Product_L48_20210604")
    cover_classes = {}
    for t, v in zip(edomv, edomvd):
        cover_classes[t.text] = v.text

    clist = [i.split() for i in root.find("eainfo/overview/eadetcit").text.split("\n")[2:]]
    colors = {
        int(c): (float(r) / 255.0, float(g) / 255.0, float(b) / 255.0, 1.0) for c, r, g, b in clist
    }
    colors[0] = (colors[0][0], colors[0][1], colors[0][2], 0.0)

    _, edomv, edomvd = _get_xml("nlcd_2019_impervious_descriptor_l48_20210604")
    descriptors = {}
    for t, v in zip(edomv, edomvd):
        tag = t.text.split(" - ")
        descriptors[tag[0]] = v.text if tag[-1].isnumeric() else f"{tag[-1]}: {v.text}"

    cyear = [2021, 2019, 2016, 2013, 2011, 2008, 2006, 2004, 2001]
    nlcd_meta = {
        "cover_years": cyear,
        "impervious_years": cyear,
        "descriptor_years": cyear,
        "canopy_years": list(range(2011, 2022)),
        "classes": cover_classes,
        "categories": {
            "Background": ("127",),
            "Water": ("11", "12"),
            "Developed": ("21", "22", "23", "24"),
            "Barren": ("31",),
            "Forest": ("41", "42", "43", "45", "46"),
            "Shrubland": ("51", "52"),
            "Herbaceous": ("71", "72", "73", "74"),
            "Planted/Cultivated": ("81", "82"),
            "Wetlands": ("90", "95"),
        },
        "descriptors": descriptors,
        "roughness": {
            "11": 0.001,
            "12": 0.022,
            "21": 0.0404,
            "22": 0.0678,
            "23": 0.0678,
            "24": 0.0404,
            "31": 0.0113,
            "41": 0.36,
            "42": 0.32,
            "43": 0.4,
            "45": 0.4,
            "46": 0.24,
            "51": 0.24,
            "52": 0.4,
            "71": 0.368,
            "81": 0.325,
            "82": 0.037,
            "90": 0.086,
            "95": 0.1825,
        },
        "colors": colors,
    }

    return nlcd_meta


def nwis_errors() -> pd.DataFrame:
    """Get error code lookup table for USGS sites that have daily values."""
    return pd.read_html(
        "https://waterservices.usgs.gov/docs/dv-service/daily-values-service-details/#error-codes"
    )[0]


def get_ssebopeta_urls(dates: tuple[str, str] | int | list[int]) -> list[tuple[pd.Timestamp, str]]:
    """Get list of URLs for SSEBop dataset within a period or years."""
    if not isinstance(dates, (tuple, list, int)):
        raise InputTypeError("dates", "tuple, list, or int", "(start, end), year, or [years, ...]")

    year = datetime.now().year - 1
    if isinstance(dates, tuple):
        if len(dates) != 2:
            raise InputTypeError("dates", "(start, end)")
        start = pd.to_datetime(dates[0])
        end = pd.to_datetime(dates[1])
        if start < pd.to_datetime("2000-01-01") or end > pd.to_datetime(f"{year}-12-31"):
            raise InputRangeError("SSEBop", ("2000", str(year)))
        date_range = pd.date_range(start, end)
    else:
        years = dates if isinstance(dates, list) else [dates]
        seebop_yrs = np.arange(2000, year)

        if any(y not in seebop_yrs for y in years):
            raise InputRangeError("SSEBop", ("2000", str(year)))

        d_list = [pd.date_range(f"{y}0101", f"{y}1231") for y in years]
        date_range = d_list.pop(0)
        while d_list:
            date = d_list.pop(0)
            date_range = date_range.union(date)  # pyright: ignore[reportOptionalMemberAccess]

    date_range = cast("pd.DatetimeIndex", date_range)
    base_url = ServiceURL().http.ssebopeta

    f_list = [
        (d, f"{base_url}/det{d.strftime('%Y%j')}.modisSSEBopETactual.zip") for d in date_range
    ]

    return f_list


@dataclass(frozen=True)
class Stats:
    """Statistics for NLCD."""

    classes: dict[str, float]
    categories: dict[str, float]


def _get_state_codes(subset_key: str | list[str]) -> list[str]:
    """Get state codes for a subset of the US."""
    keys = [subset_key] if isinstance(subset_key, str) else subset_key
    state_cd = []

    state_keys = [k.upper() for k in keys if len(k) == 2]
    states = us_abbrs.STATES
    if any(k not in states for k in state_keys):
        raise InputValueError("subset_key", states)
    if state_keys:
        state_cd += state_keys

    other_keys = [k for k in keys if len(k) > 2]
    if "conus" in other_keys:
        other_keys.remove("conus")
        other_keys.append("contiguous")
    valid_keys = ["contiguous", "continental", "territories", "commonwealths"]
    if any(k not in valid_keys for k in other_keys):
        raise InputValueError("subset_key", [*valid_keys, "conus"])
    if other_keys:
        state_cd += tlz.concat(getattr(us_abbrs, k.upper()) for k in other_keys)
    return state_cd


def get_us_states(subset_key: str | list[str] | None = None) -> gpd.GeoDataFrame:
    """Get US states as a GeoDataFrame from Census' TIGERLine 2023 database.

    Parameters
    ----------
    subset_key : str or list of str, optional
        Key to subset the geometries instead of returning all states, by default
        all states are returned. Valid keys are:

        - ``contiguous`` or ``conus``
        - ``continental``
        - ``commonwealths``
        - ``territories``
        - Two letter state codes, e.g., ``["TX", "CA", "FL", ...]``

    Returns
    -------
    geopandas.GeoDataFrame
        GeoDataFrame of requested US states.
    """
    url = "https://www2.census.gov/geo/tiger/TIGER2023/STATE/tl_2023_us_state.zip"
    us_states = gpd.read_file(io.BytesIO(ar.retrieve_binary([url])[0]))
    if subset_key is not None:
        state_cd = _get_state_codes(subset_key)
        return us_states[us_states.STUSPS.isin(state_cd)].copy()  # pyright: ignore[reportReturnType]
    return us_states  # pyright: ignore[reportReturnType]


@dataclass(frozen=True)
class StateCounties:
    """State and county codes and names."""

    name: str
    code: str | None
    counties: pd.Series


def states_lookup_table() -> dict[str, StateCounties]:
    """Get codes and names of US states and their counties.

    Notes
    -----
    This function is based on a file prepared by developers of
    an R package called `dataRetrieval <https://github.com/USGS-R/dataRetrieval>`__.

    Returns
    -------
    pandas.DataFrame
        State codes and name as a dataframe.
    """
    urls = [
        "https://www2.census.gov/geo/docs/reference/state.txt",
        "/".join(
            [
                "https://raw.githubusercontent.com/USGS-R/dataRetrieval",
                "main/inst/extdata/state_county.json",
            ]
        ),
    ]
    resp = ar.retrieve_text(urls)

    codes = pd.read_csv(io.StringIO(resp[0]), sep="|")
    codes["STATE"] = codes["STATE"].astype(str).str.zfill(2)
    codes = codes.set_index("STATE")

    def _county2series(cd: dict[str, dict[str, str]]) -> pd.Series:
        return pd.DataFrame.from_dict(cd, orient="index")["name"]  # pyright: ignore[reportReturnType]

    def _state_cd(state: str) -> str | None:
        try:
            return codes.loc[state, "STUSAB"]
        except KeyError:
            return None

    states = {
        c: StateCounties(s["name"], _state_cd(c), _county2series(s["county_cd"]))
        for c, s in json.loads(resp[1])["US"]["state_cd"].items()
    }
    return states



"""Accessing National Flood Hazard Layers (NFHL) through web services."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

from pygeohydro.exceptions import InputValueError
from pygeoogc import ServiceURL
from pynhd import AGRBase

if TYPE_CHECKING:
    import pyproj

    CRSTYPE = Union[int, str, pyproj.CRS]

__all__ = ["NFHL"]


class NFHL(AGRBase):
    """Access National Flood Hazard Layers (NFHL).

    Parameters
    ----------
    service : str
        The service type. Valid services are:

        - ``NFHL``: Effective National Flood Hazard Layers
        - ``Prelim_CSLF``: Preliminary Changes Since Last Firm (CSLF)
        - ``Draft_CSLF``: Draft Changes Since Last Firm (CSLF)
        - ``Prelim_NFHL``: Preliminary National Flood Hazard Layers
        - ``Pending_NFHL``: Pending National Flood Hazard Layers
        - ``Draft_NFHL``: Draft National Flood Hazard Layers

    layer : str
        A valid service layer. Valid layers are service specific:

        - ``NFHL``: ``nfhl availability``, ``firm panels``, ``lomrs``, ``lomas``,
            ``political jurisdictions``, ``profile baselines``, ``water lines``,
            ``cross-sections``, ``base flood elevations``, ``levees``,
            ``seclusion boundaries``, ``coastal transects``, ``transect baselines``,
            ``general structures``, ``river mile markers``, ``water areas``, ``plss``,
            ``limit of moderate wave action``, ``flood hazard boundaries``,
            ``flood hazard zones``, ``primary frontal dunes``, ``base index``,
            ``topographic low confidence areas``, ``datum conversion points``,
            ``coastal gages``, ``gages``, ``nodes``, ``high water marks``,
            ``station start points``, ``hydrologic reaches``, ``alluvial fans``,
            and ``subbasins``
        - ``Prelim_CSLF``: ``preliminary``, ``coastal high hazard area change``,
            ``floodway change``, ``special flood hazard area change``,
            and ``non-special flood hazard area change``
        - ``Draft_CSLF``: ``draft``, ``coastal high hazard area change``,
            ``floodway change``, ``special flood hazard area change``, and
            ``non-special flood hazard area change``
        - ``Prelim_NFHL``: ``preliminary data availability``,
            ``preliminary firm panel index``, ``preliminary plss``,
            ``preliminary topographic low confidence areas``,
            ``preliminary river mile markers``, ``preliminary datum conversion points``,
            ``preliminary coastal gages``, ``preliminary gages``, ``preliminary nodes``,
            ``preliminary high water marks``, ``preliminary station start points``,
            ``preliminary cross-sections``, ``preliminary coastal transects``,
            ``preliminary base flood elevations``, ``preliminary profile baselines``,
            ``preliminary transect baselines``, ``preliminary limit of moderate wave action``,
            ``preliminary water lines``, ``preliminary political jurisdictions``,
            ``preliminary levees``, ``preliminary general structures``,
            ``preliminary primary frontal dunes``, ``preliminary hydrologic reaches``,
            ``preliminary flood hazard boundaries``, ``preliminary flood hazard zones``,
            ``preliminary submittal information``, ``preliminary alluvial fans``,
            ``preliminary subbasins``, and ``preliminary water areas``
        - ``Pending_NFHL``: ``pending submittal information``, ``pending water areas``,
            ``pending firm panel index``, ``pending data availability``,
            ``pending firm panels``, ``pending political jurisdictions``,
            ``pending profile baselines``, ``pending water lines``,
            ``pending cross-sections``, ``pending base flood elevations``,
            ``pending levees``, ``pending seclusion boundaries``,
            ``pending coastal transects``, ``pending transect baselines``,
            ``pending general structures``, ``pending river mile markers``,
            ``pending plss``, ``pending limit of moderate wave action``,
            ``pending flood hazard boundaries``, ``pending flood hazard zones``,
            ``pending primary frontal dunes``, ``pending topographic low confidence areas``,
            ``pending datum conversion points``, ``pending coastal gages``,
            ``pending gages``, ``pending nodes``, ``pending high water marks``,
            ``pending station start points``, ``pending hydrologic reaches``,
            ``pending alluvial fans``, and ``pending subbasins``
        - ``Draft_NFHL``: ``draft data availability``, ``draft firm panels``,
            ``draft political jurisdictions``, ``draft profile baselines``,
            ``draft water lines``, ``draft cross-sections``, ``draft base flood elevations``,
            ``draft levees``, ``draft submittal info``, ``draft coastal transects``,
            ``draft transect baselines``, ``draft general structures``,
            ``draft limit of moderate wave action``, ``draft flood hazard boundaries``,
            and ``draft flood hazard zones``

    outfields : str or list, optional
        Target field name(s), default to "*" i.e., all the fields.
    crs : str, int, or pyproj.CRS, optional
        Target spatial reference of output, default to ``EPSG:4326``.

    Examples
    --------
    >>> from pygeohydro import NFHL
    >>> nfhl = NFHL("NFHL", "cross-sections")
    >>> gdf_xs = nfhl.bygeom((-73.42, 43.28, -72.9, 43.52), geo_crs=4269)

    References
    ----------
    * `National Flood Hazard Layer <https://hazards.fema.gov/femaportal/wps/portal/NFHLWMS>`__

    Methods
    -------
    bygeom(geom, geo_crs=4326, sql_clause="", distance=None, return_m=False, return_geom=True)
        Get features within a geometry that can be combined with a SQL where clause.
    byids(field, fids, return_m=False, return_geom=True)
        Get features by object IDs.
    bysql(sql_clause, return_m=False, return_geom=True)
        Get features using a valid SQL 92 WHERE clause.
    """

    def __init__(
        self, service: str, layer: str, outfields: str | list[str] = "*", crs: CRSTYPE = 4326
    ):
        # service URLs
        self.__valid_services = {
            "NFHL": ServiceURL().restful.fema_nfhl,
            "Prelim_CSLF": ServiceURL().restful.fema_prelim_cslf,
            "Draft_CSLF": ServiceURL().restful.fema_draft_cslf,
            "Prelim_NFHL": ServiceURL().restful.fema_prelim_nfhl,
            "Pending_NFHL": ServiceURL().restful.fema_pending_nfhl,
            "Draft_NFHL": ServiceURL().restful.fema_draft_nfhl,
        }
        url = self.valid_services.get(service)
        if url is None:
            raise InputValueError("service", list(self.valid_services))

        super().__init__(url, layer, outfields, crs)

    @property
    def valid_services(self) -> dict[str, str]:
        """A dictionary of valid services and their URLs."""
        return self.__valid_services



"""Accessing data from the supported databases through their APIs."""

from __future__ import annotations

import importlib.util
import warnings
from datetime import datetime, timezone
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable, Sequence, Tuple, Union, cast

import geopandas as gpd
import pandas as pd
import requests

import async_retriever as ar
import pygeoogc as ogc
import pygeoutils as geoutils
from pygeohydro.exceptions import (
    InputTypeError,
    InputValueError,
    ServiceError,
    ZeroMatchedError,
)
from pygeoogc import ServiceURL
from pygeoutils import EmptyResponseError

if TYPE_CHECKING:
    import pyproj
    from shapely import MultiPolygon, Polygon

    GTYPE = Union[Polygon, MultiPolygon, Tuple[float, float, float, float]]
    CRSTYPE = Union[int, str, pyproj.CRS]

__all__ = ["NID"]


def _remote_file_modified(file_path: Path) -> bool:
    """Check if the file is older than the last modification date of the NID web service."""
    if not file_path.exists():
        return True
    url = "https://nid.sec.usace.army.mil/api/nation/gpkg"
    # we need to get the redirect URL so that we can get the last modified date, so
    # we need to send a request with the Range header set to 0-0 to avoid downloading
    # the entire file.
    response = requests.get(url, headers={"Range": "bytes=0-0"}, allow_redirects=True, timeout=50)
    if response.status_code not in (200, 206):
        raise ServiceError(response.reason, url)
    response = requests.head(response.url, timeout=50)
    if response.status_code != 200:
        raise ServiceError(response.reason, url)

    remote_last_modified = datetime.strptime(
        response.headers["Last-Modified"], "%a, %d %b %Y %H:%M:%S GMT"
    ).replace(tzinfo=timezone.utc)
    local_last_modified = datetime.fromtimestamp(file_path.stat().st_mtime, tz=timezone.utc)
    return local_last_modified < remote_last_modified


class NID:
    """Retrieve data from the National Inventory of Dams web service."""

    def __init__(self) -> None:
        self.base_url = ServiceURL().restful.nid
        self.suggest_url = f"{self.base_url}/suggestions"
        resp = ar.retrieve_json([f"{self.base_url}/advanced-fields"])
        resp = cast("list[dict[str, Any]]", resp)
        self.fields_meta = pd.DataFrame(resp[0])
        self.valid_fields = self.fields_meta["name"]
        self.dam_type = {
            pd.NA: "N/A",
            None: "N/A",
            1: "Arch",
            2: "Multi-Arch",
            3: "Stone",
            4: "Roller-Compacted Concrete",
            5: "Rockfill",
            6: "Buttress",
            7: "Masonry",
            8: "Earth",
            9: "Gravity",
            10: "Timber Crib",
            11: "Concrete",
            12: "Other",
        }
        dtype_str = {str(k): v for k, v in self.dam_type.items() if str(k).isdigit()}
        self.dam_type = {**self.dam_type, **dtype_str}
        self.dam_purpose = {
            pd.NA: "N/A",
            None: "N/A",
            1: "Tailings",
            2: "Irrigation",
            3: "Navigation",
            4: "Fish and Wildlife Pond",
            5: "Recreation",
            6: "Hydroelectric",
            7: "Debris Control",
            8: "Water Supply",
            9: "Flood Risk Reduction",
            10: "Fire Protection, Stock, Or Small Farm Pond",
            11: "Grade Stabilization",
            12: "Other",
        }
        purp_str = {str(k): v for k, v in self.dam_purpose.items() if str(k).isdigit()}
        self.dam_purpose = {**self.dam_purpose, **purp_str}
        self.data_units = {
            "distance": "mile",
            "damHeight": "ft",
            "hydraulicHeight": "ft",
            "structuralHeight": "ft",
            "nidHeight": "ft",
            "damLength": "ft",
            "volume": "cubic yards",
            "nidStorage": "acre-ft",
            "maxStorage": "acre-ft",
            "normalStorage": "acre-ft",
            "surfaceArea": "acre",
            "drainageArea": "square miles",
            "maxDischarge": "cfs",
            "spillwayWidth": "ft",
        }
        self._nid_inventory_path = Path("cache", "full_nid_inventory.parquet")

    @property
    def nid_inventory_path(self) -> Path:
        """Path to the NID inventory parquet file."""
        return self._nid_inventory_path

    @nid_inventory_path.setter
    def nid_inventory_path(self, value: Path | str) -> None:
        self._nid_inventory_path = Path(value)
        self._nid_inventory_path.parent.mkdir(parents=True, exist_ok=True)

    def stage_nid_inventory(self, fname: str | Path | None = None) -> None:
        """Download the entire NID inventory data and save to a parquet file.

        Parameters
        ----------
        fname : str, pathlib.Path, optional
            The path to the file to save the data to, defaults to
            ``./cache/full_nid_inventory.parquet``.
        """
        fname = self.nid_inventory_path if fname is None else Path(fname)
        if fname.suffix != ".parquet":
            fname = fname.with_suffix(".parquet")

        self.nid_inventory_path = fname
        gpkg_file = fname.with_suffix(".gpkg")
        if _remote_file_modified(gpkg_file) or not self.nid_inventory_path.exists():
            gpkg_file.unlink(missing_ok=True)
            url = "https://nid.sec.usace.army.mil/api/nation/gpkg"
            fname_ = ogc.streaming_download(url, fnames=gpkg_file)
            if fname_ is None:
                raise EmptyResponseError
            dams = (
                gpd.read_file(gpkg_file, engine="pyogrio", use_arrow=True)
                if importlib.util.find_spec("pyogrio")
                else gpd.read_file(gpkg_file)
            )
            dams = cast("gpd.GeoDataFrame", dams)

            dams = dams.astype(
                {
                    "name": str,
                    "otherNames": str,
                    "formerNames": str,
                    "nidId": str,
                    "otherStructureId": str,
                    "federalId": str,
                    "ownerNames": str,
                    "ownerTypeIds": str,
                    "primaryOwnerTypeId": str,
                    "separateStructuresCount": str,
                    "isAssociatedStructureId": str,
                    "designerNames": str,
                    "nonFederalDamOnFederalId": str,
                    "primaryPurposeId": str,
                    "purposeIds": str,
                    "sourceAgency": str,
                    "stateFedId": str,
                    "latitude": "f8",
                    "longitude": "f8",
                    "state": str,
                    "county": str,
                    "countyState": str,
                    "city": str,
                    "distance": "f8",
                    "riverName": str,
                    "congDist": str,
                    "stateRegulatedId": str,
                    "jurisdictionAuthorityId": str,
                    "stateRegulatoryAgency": str,
                    "permittingAuthorityId": str,
                    "inspectionAuthorityId": str,
                    "enforcementAuthorityId": str,
                    "fedRegulatedId": str,
                    "fedOwnerIds": str,
                    "fedFundingIds": str,
                    "fedDesignIds": str,
                    "fedConstructionIds": str,
                    "fedRegulatoryIds": str,
                    "fedInspectionIds": str,
                    "fedOperationIds": str,
                    "fedOtherIds": str,
                    "secretaryAgricultureBuiltId": str,
                    "nrcsWatershedAuthorizationId": str,
                    "primaryDamTypeId": str,
                    "damTypeIds": str,
                    "coreTypeIds": str,
                    "foundationTypeIds": str,
                    "damHeight": "f8",
                    "hydraulicHeight": "f8",
                    "structuralHeight": "f8",
                    "nidHeight": "f8",
                    "nidHeightId": str,
                    "damLength": "f8",
                    "volume": "f8",
                    "yearCompleted": "Int32",
                    "yearCompletedId": str,
                    "yearsModified": str,
                    "nidStorage": "f8",
                    "maxStorage": "f8",
                    "normalStorage": "f8",
                    "surfaceArea": "f8",
                    "drainageArea": "f8",
                    "maxDischarge": "f8",
                    "spillwayTypeId": str,
                    "spillwayWidth": "f8",
                    "numberOfLocks": "Int32",
                    "lengthOfLocks": "f8",
                    "widthOfLocks": "f8",
                    "secondaryLengthOfLocks": "Int32",
                    "secondaryWidthOfLocks": "Int32",
                    "outletGateTypes": str,
                    "dataUpdated": "datetime64[ns]",
                    "inspectionDate": str,
                    "inspectionFrequency": "f4",
                    "hazardId": str,
                    "conditionAssessId": str,
                    "conditionAssessDate": "datetime64[ns]",
                    "eapId": str,
                    "eapLastRevDate": "datetime64[ns]",
                    "websiteUrl": str,
                    "usaceDivision": str,
                    "usaceDistrict": str,
                    "operationalStatusId": str,
                    "operationalStatusDate": "datetime64[ms]",
                    "inundationNidAddedId": str,
                    "huc2": str,
                    "huc4": str,
                    "huc6": str,
                    "huc8": str,
                    "zipcode": str,
                    "nation": str,
                    "stateKey": str,
                    "femaRegion": str,
                    "femaCommunity": str,
                    "aiannh": str,
                }
            )
            for c in dams:
                if (dams[c] == "Yes").any():
                    dams[c] = dams[c] == "Yes"
            dams.loc[dams["yearCompleted"] < 1000, "yearCompleted"] = pd.NA
            dams.to_parquet(fname)

    @property
    def df(self):
        """Entire NID inventory (``csv`` version) as a ``pandas.DataFrame``."""
        fname = self.nid_inventory_path
        par_name = fname.with_suffix(".parquert")
        if par_name.exists():
            return pd.read_parquet(par_name)
        url = "https://nid.sec.usace.army.mil/api/nation/csv"
        fname = ogc.streaming_download(url, fnames=fname.with_suffix(".csv"))
        if fname is None:
            raise EmptyResponseError
        dams = pd.read_csv(fname, header=1, engine="pyarrow")
        dams.to_parquet(par_name)
        return dams

    @property
    def gdf(self):
        """Entire NID inventory (``gpkg`` version) as a ``geopandas.GeoDataFrame``."""
        self.stage_nid_inventory()
        return gpd.read_parquet(self.nid_inventory_path)

    @staticmethod
    def _get_json(
        urls: Sequence[str], params: list[dict[str, str]] | None = None
    ) -> list[dict[str, Any]]:
        """Get JSON response from NID web service.

        Parameters
        ----------
        urls : list of str
            A list of query URLs.
        params : dict, optional
            A list of parameters to pass to the web service, defaults to ``None``.

        Returns
        -------
        list of dict
            List of JSON responses from the web service.
        """
        if not isinstance(urls, list):
            raise InputTypeError("urls", "list or str")

        kwds = None if params is None else [{"params": {**p, "out": "json"}} for p in params]
        resp = ar.retrieve_json(urls, kwds)
        resp = cast("list[dict[str, Any]]", resp)
        if not resp:
            raise ZeroMatchedError

        failed = [(i, f"Req_{i}: {r['message']}") for i, r in enumerate(resp) if "error" in r]
        if failed:
            idx, err_msgs = zip(*failed)
            idx = cast("Tuple[int]", idx)
            err_msgs = cast("Tuple[str]", err_msgs)
            errs = " service requests failed with the following messages:\n"
            errs += "\n".join(err_msgs)
            if len(failed) == len(urls):
                raise ZeroMatchedError(f"All{errs}")
            resp = [r for i, r in enumerate(resp) if i not in idx]
            fail_count = f"{len(failed)} of {len(urls)}"
            warnings.warn(f"{fail_count}{errs}", UserWarning, stacklevel=2)
        return resp

    @staticmethod
    def _to_geodf(nid_df: pd.DataFrame) -> gpd.GeoDataFrame:
        """Convert a NID dataframe to a GeoDataFrame.

        Parameters
        ----------
        dams : pd.DataFrame
            NID dataframe

        Returns
        -------
        geopandas.GeoDataFrame
            GeoDataFrame of NID data
        """
        return gpd.GeoDataFrame(  # pyright: ignore[reportCallIssue]
            nid_df,
            geometry=gpd.points_from_xy(nid_df["longitude"], nid_df["latitude"], crs=4326),
        )

    def get_byfilter(self, query_list: list[dict[str, list[str]]]) -> list[gpd.GeoDataFrame]:
        """Query dams by filters from the National Inventory of Dams web service.

        Parameters
        ----------
        query_list : list of dict
            List of dictionary of query parameters. For an exhaustive list of the parameters,
            use the advanced fields dataframe that can be accessed via ``NID().fields_meta``.
            Some filter require min/max values such as ``damHeight`` and ``drainageArea``.
            For such filters, the min/max values should be passed like so:
            ``{filter_key: ["[min1 max1]", "[min2 max2]"]}``.

        Returns
        -------
        list of geopandas.GeoDataFrame
            Query results in the same order as the input query list.

        Examples
        --------
        >>> from pygeohydro import NID
        >>> nid = NID()
        >>> query_list = [
        ...    {"drainageArea": ["[200 500]"]},
        ...    {"nidId": ["CA01222"]},
        ... ]
        >>> dam_dfs = nid.get_byfilter(query_list)
        """
        fields = self.valid_fields.to_list()
        invalid = [k for key in query_list for k in key if k not in fields]
        if invalid:
            raise InputValueError("query_dict", fields)
        params = [
            {"sy": " ".join(f"@{s}:{fid}" for s, fids in key.items() for fid in fids)}
            for key in query_list
        ]
        return [
            self._to_geodf(pd.DataFrame(r))
            for r in self._get_json([f"{self.base_url}/query"] * len(params), params)
        ]

    def get_bygeom(self, geometry: GTYPE, geo_crs: CRSTYPE) -> gpd.GeoDataFrame:
        """Retrieve NID data within a geometry.

        Parameters
        ----------
        geometry : Polygon, MultiPolygon, or tuple of length 4
            Geometry or bounding box (west, south, east, north) for extracting the data.
        geo_crs : list of str
            The CRS of the input geometry.

        Returns
        -------
        geopandas.GeoDataFrame
            GeoDataFrame of NID data

        Examples
        --------
        >>> from pygeohydro import NID
        >>> nid = NID()
        >>> dams = nid.get_bygeom((-69.77, 45.07, -69.31, 45.45), 4326)
        """
        _geometry = geoutils.geo2polygon(geometry, geo_crs, self.gdf.crs)
        idx = self.gdf.sindex.query(_geometry, "contains")
        return self.gdf.iloc[idx].copy()

    def inventory_byid(self, federal_ids: list[str]) -> gpd.GeoDataFrame:
        """Get extra attributes for dams based on their dam ID.

        Notes
        -----
        This function is meant to be used for getting extra attributes for dams.
        For example, first you need to use either ``get_bygeom`` or ``get_byfilter``
        to get basic attributes of the target dams. Then you can use this function
        to get extra attributes using the ``id`` column of the ``GeoDataFrame``
        that ``get_bygeom`` or ``get_byfilter`` returns.

        Parameters
        ----------
        federal_ids : list of str
            List of the target dam Federal IDs.

        Returns
        -------
        pandas.DataFrame
            Dams with extra attributes in addition to the standard NID fields
            that other ``NID`` methods return.

        Examples
        --------
        >>> from pygeohydro import NID
        >>> nid = NID()
        >>> dams = nid.inventory_byid(['KY01232', 'GA02400', 'NE04081', 'IL55070', 'TN05345'])
        """
        if not isinstance(federal_ids, Iterable) or isinstance(federal_ids, (str, int)):
            raise InputTypeError("federal_ids", "list of str (Federal IDs)")

        if not all(isinstance(i, str) for i in federal_ids):
            raise InputTypeError("federal_ids", "list of str (Federal IDs)")

        urls = [f"{self.base_url}/dams/{i.upper()}/inventory" for i in set(federal_ids)]
        return self._to_geodf(pd.DataFrame(self._get_json(urls)).set_index("id"))

    def get_suggestions(
        self, text: str, context_key: str | None = None
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Get suggestions from the National Inventory of Dams web service.

        Notes
        -----
        This function is useful for exploring and/or narrowing down the filter fields
        that are needed to query the dams using ``get_byfilter``.

        Parameters
        ----------
        text : str
            Text to query for suggestions.
        context_key : str, optional
            Suggestion context, defaults to empty string, i.e., all context keys.
            For a list of valid context keys, see ``NID().fields_meta``.

        Returns
        -------
        tuple of pandas.DataFrame
            The suggestions for the requested text as two DataFrames:
            First, is suggestions found in the dams properties and
            second, those found in the query fields such as states, huc6, etc.

        Examples
        --------
        >>> from pygeohydro import NID
        >>> nid = NID()
        >>> dams, contexts = nid.get_suggestions("houston", "city")
        """
        fields = self.valid_fields.to_list()
        params = {"text": text}
        if context_key:
            if context_key not in fields:
                raise InputValueError("context", fields)
            params["contextKey"] = context_key
        resp = self._get_json([f"{self.base_url}/suggestions"], [params])
        dams = pd.DataFrame(resp[0]["dams"])
        contexts = pd.DataFrame(resp[0]["contexts"])
        return (
            dams if dams.empty else dams.set_index("id"),
            contexts if contexts.empty else contexts.set_index("name"),
        )

    def __repr__(self) -> str:
        """Print the services properties."""
        resp = self._get_json([f"{self.base_url}/metadata"])[0]
        return "\n".join(
            [
                "NID RESTful information:",
                f"URL: {self.base_url}",
                f"Date Refreshed: {resp['dateRefreshed']}",
                f"Version: {resp['version']}",
            ]
        )



"""Accessing watershed boundary-level data through web services."""

from __future__ import annotations

import importlib.util
import io
from pathlib import Path
from typing import TYPE_CHECKING, Union, cast

import geopandas as gpd
import pandas as pd
import xarray as xr

import async_retriever as ar
import pygeoogc as ogc
from pygeohydro.exceptions import InputValueError
from pygeoogc import ServiceURL
from pynhd import AGRBase
from pynhd.core import ScienceBase

if TYPE_CHECKING:
    import pyproj

    CRSTYPE = Union[int, str, pyproj.CRS]

__all__ = [
    "WBD",
    "huc_wb_full",
    "irrigation_withdrawals",
]


class WBD(AGRBase):
    """Access Watershed Boundary Dataset (WBD).

    Notes
    -----
    This web service offers Hydrologic Unit (HU) polygon boundaries for
    the United States, Puerto Rico, and the U.S. Virgin Islands.
    For more info visit: https://hydro.nationalmap.gov/arcgis/rest/services/wbd/MapServer

    Parameters
    ----------
    layer : str, optional
        A valid service layer. Valid layers are:

        - ``wbdline``
        - ``huc2``
        - ``huc4``
        - ``huc6``
        - ``huc8``
        - ``huc10``
        - ``huc12``
        - ``huc14``
        - ``huc16``

    outfields : str or list, optional
        Target field name(s), default to "*" i.e., all the fields.
    crs : str, int, or pyproj.CRS, optional
        Target spatial reference, default to ``EPSG:4326``.
    """

    def __init__(self, layer: str, outfields: str | list[str] = "*", crs: CRSTYPE = 4326):
        self.valid_layers = {
            "wbdline": "wbdline",
            "huc2": "2-digit hu (region)",
            "huc4": "4-digit hu (subregion)",
            "huc6": "6-digit hu (basin)",
            "huc8": "8-digit hu  (subbasin)",
            "huc10": "10-digit hu (watershed)",
            "huc12": "12-digit hu (subwatershed)",
            "huc14": "14-digit hu",
            "huc16": "16-digit hu",
        }
        _layer = self.valid_layers.get(layer)
        if _layer is None:
            raise InputValueError("layer", list(self.valid_layers))
        super().__init__(ServiceURL().restful.wbd, _layer, outfields, crs)


def huc_wb_full(huc_lvl: int) -> gpd.GeoDataFrame:
    """Get the full watershed boundary for a given HUC level.

    Notes
    -----
    This function is designed for cases where the full watershed boundary is needed
    for a given HUC level. If only a subset of the HUCs is needed, then use
    the ``pygeohydro.WBD`` class. The full dataset is downloaded from the National Maps'
    `WBD staged products <https://prd-tnm.s3.amazonaws.com/index.html?prefix=StagedProducts/Hydrography/WBD/HU2/Shape/>`__.

    Parameters
    ----------
    huc_lvl : int
        HUC level, must be even numbers between 2 and 16.

    Returns
    -------
    geopandas.GeoDataFrame
        The full watershed boundary for the given HUC level.
    """
    valid_hucs = [2, 4, 6, 8, 10, 12, 14, 16]
    if huc_lvl not in valid_hucs:
        raise InputValueError("huc_lvl", list(map(str, valid_hucs)))

    base_url = "https://prd-tnm.s3.amazonaws.com/StagedProducts/Hydrography/WBD/HU2/Shape"

    urls = [f"{base_url}/WBD_{h2:02}_HU2_Shape.zip" for h2 in range(1, 23)]
    fnames = [Path("cache", Path(url).name) for url in urls]
    fnames = ogc.streaming_download(urls, fnames=fnames)
    fnames = [f for f in fnames if f is not None]
    keys = (p.stem.split("_")[1] for p in fnames)
    engine = "pyogrio" if importlib.util.find_spec("pyogrio") else "fiona"
    huc = (
        gpd.GeoDataFrame(
            pd.concat(
                (gpd.read_file(f"{p}!Shape/WBDHU{huc_lvl}.shp", engine=engine) for p in fnames),
                keys=keys,
            )
        )
        .reset_index()
        .rename(columns={"level_0": "huc2"})
        .drop(columns="level_1")
    )
    huc = cast("gpd.GeoDataFrame", huc)
    return huc


def irrigation_withdrawals() -> xr.Dataset:
    """Get monthly water use for irrigation at HUC12-level for CONUS.

    Notes
    -----
    Dataset is retrieved from https://doi.org/10.5066/P9FDLY8P.
    """
    item = ScienceBase.get_file_urls("5ff7acf4d34ea5387df03d73")
    urls = item.loc[item.index.str.contains(".csv"), "url"]
    resp = ar.retrieve_text(urls.tolist())
    irr = {}
    for name, r in zip(urls.index, resp):
        df = pd.read_csv(
            io.StringIO(r),
            usecols=lambda s: "m3" in s or "huc12t" in s,  # type: ignore
        )
        df["huc12t"] = df["huc12t"].str.strip("'")
        df = df.rename(columns={"huc12t": "huc12"}).set_index("huc12")
        df = df.rename(columns={c: str(c)[:3].capitalize() for c in df})
        irr[name[-6:-4]] = df.copy()
    ds = xr.Dataset(irr).rename({"dim_1": "month"})
    long_names = {
        "GW": "groundwater_withdrawal",
        "SW": "surface_water_withdrawal",
        "TW": "total_withdrawal",
        "CU": "consumptive_use",
    }
    for v, n in long_names.items():
        ds[v].attrs["long_name"] = n
        ds[v].attrs["units"] = "m3"
    ds.attrs["description"] = " ".join(
        (
            "Estimated Monthly Water Use for Irrigation by",
            "12-Digit Hydrologic Unit in the Conterminous United States for 2015",
        )
    )
    ds.attrs["source"] = "https://doi.org/10.5066/P9FDLY8P"
    return ds



"""Access USGS Short-Term Network (STN) via Restful API."""

from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING, Any, ClassVar, Literal, Union, cast, overload

import geopandas as gpd
import numpy as np
import pandas as pd

import async_retriever as ar
from pygeohydro.exceptions import InputValueError
from pygeoogc import ServiceURL

if TYPE_CHECKING:
    from pyproj import CRS

    CRSTYPE = Union[int, str, CRS]

__all__ = ["STNFloodEventData", "stn_flood_event"]


class STNFloodEventData:
    """Client for STN Flood Event Data's RESTFUL Service API.

    Advantages of using this client are:

    - The user does not need to know the details of RESTFUL in
      general and of this API specifically.
    - Parses the data and returns Python objects
      (e.g., pandas.DataFrame, geopandas.GeoDataFrame) instead of JSON.
    - Convenience functions are offered for data dictionaries.
    - Geo-references the data where applicable.

    Attributes
    ----------
    service_url : str
        The service url of the STN Flood Event Data RESTFUL Service API.
    data_dictionary_url : str
        The data dictionary url of the STN Flood Event Data RESTFUL Service API.
    service_crs : int
        The CRS of the data from the service which is ``EPSG:4326``.
    instruments_query_params : set
        The accepted query parameters for the instruments data type.
        Accepted values are ``SensorType``, ``CurrentStatus``, ``States``,
        ``Event``, ``County``, ``DeploymentType``, ``EventType``,
        ``EventStatus``, and ``CollectionCondition``.
    peaks_query_params : set
        The accepted query parameters for the peaks data type.
        Accepted values are ``EndDate``, ``States``, ``Event``, ``StartDate``,
        ``County``, ``EventType``, and ``EventStatus``.
    hwms_query_params : set
        The accepted query parameters for the hwms data type.
        Accepted values are ``EndDate``, ``States``, ``Event``, ``StartDate``,
        ``County``, ``EventType``, and ``EventStatus``.
    sites_query_params : set
        The accepted query parameters for the sites data type.
        Accepted values are ``OPDefined``, ``HousingTypeOne``, ``NetworkName``,
        ``HousingTypeSeven``, ``RDGOnly``, ``HWMOnly``, ``Event``,
        ``SensorOnly``, ``State``, ``SensorType``, and ``HWMSurveyed``.


    Notes
    -----
    Point data from the service is assumed to be in the WGS84
    coordinate reference system (``EPSG:4326``).

    References
    ----------
    * `USGS Short-Term Network (STN) <https://stn.wim.usgs.gov/STNWeb/#/>`__
    * `All Sensors API Documentation <https://stn.wim.usgs.gov/STNServices/Documentation/Sensor/AllSensors>`__
    * `All Peak Summary API Documentation <https://stn.wim.usgs.gov/STNServices/Documentation/PeakSummary/AllPeakSummaries>`__
    * `All HWM API Documentation <https://stn.wim.usgs.gov/STNServices/Documentation/HWM/AllHWMs>`__
    * `All Sites API Documentation <https://stn.wim.usgs.gov/STNServices/Documentation/Site/AllSites>`__
    * `USGS Flood Event Viewer: Providing Hurricane and Flood Response Data <https://www.usgs.gov/mission-areas/water-resources/science/usgs-flood-event-viewer-providing-hurricane-and-flood>`__
    * `A USGS guide for finding and interpreting high-water marks <https://www.usgs.gov/media/videos/a-usgs-guide-finding-and-interpreting-high-water-marks>`__
    * `High-Water Marks and Flooding <https://www.usgs.gov/special-topics/water-science-school/science/high-water-marks-and-flooding>`__
    * `Identifying and preserving high-water mark data <https://doi.org/10.3133/tm3A24>`__
    """

    # Per Athena Clark, Lauren Privette, and Hans Vargas at USGS
    # this is the CRS used for visualization on STN front-end.
    service_crs: ClassVar[int] = 4326
    service_url: ClassVar[str] = ServiceURL().restful.stnflood
    data_dictionary_url: ClassVar[str] = ServiceURL().restful.stnflood_dd
    instruments_query_params: ClassVar[set[str]] = {
        "Event",
        "EventType",
        "EventStatus",
        "States",
        "County",
        "CurrentStatus",
        "CollectionCondition",
        "SensorType",
        "DeploymentType",
    }
    peaks_query_params: ClassVar[set[str]] = {
        "Event",
        "EventType",
        "EventStatus",
        "States",
        "County",
        "StartDate",
        "EndDate",
    }
    hwms_query_params: ClassVar[set[str]] = {
        "Event",
        "EventType",
        "EventStatus",
        "States",
        "County",
        "StartDate",
        "EndDate",
    }
    sites_query_params: ClassVar[set[str]] = {
        "Event",
        "State",
        "SensorType",
        "NetworkName",
        "OPDefined",
        "HWMOnly",
        "HWMSurveyed",
        "SensorOnly",
        "RDGOnly",
        "HousingTypeOne",
        "HousingTypeSeven",
    }

    @staticmethod
    def _geopandify(
        input_list: list[dict[str, Any]],
        x_col: str,
        y_col: str,
        crs: CRSTYPE | None,
        service_crs: CRSTYPE,
    ) -> gpd.GeoDataFrame:
        """Georeference a list of dictionaries to a GeoDataFrame.

        Parameters
        ----------
        input_list : list of dict
            The list of dictionaries to be converted to a geodataframe.
        x_col : str
            The name of the column containing the x-coordinate.
        y_col : str
            The name of the column containing the y-coordinate.
        crs : int, str, or pyproj.CRS
            Desired Coordinate reference system (CRS) of output.
            Only used for GeoDataFrames outputs.
        service_crs : int, str, or pyproj.CRS
            The coordinate reference system of the data from the service.

        Returns
        -------
        geopandas.GeoDataFrame
            The geo-referenced GeoDataFrame.
        """
        df = pd.DataFrame(input_list)

        if crs is None:
            crs = service_crs

        gdf = gpd.GeoDataFrame(
            df,
            geometry=gpd.points_from_xy(df[x_col], df[y_col], crs=service_crs),
        )
        gdf = cast("gpd.GeoDataFrame", gdf.to_crs(crs))
        return gdf

    @staticmethod
    def _delist_dict(d: dict[str, list[float] | float]) -> dict[str, float]:
        """De-lists all unit length lists in a dictionary."""

        def delist(x: list[float] | float) -> float:
            if not isinstance(x, list):
                return x
            if len(x) == 1:
                return x[0]
            return np.nan

        return {k: delist(v) for k, v in d.items()}

    @classmethod
    @overload
    def data_dictionary(
        cls,
        data_type: str,
        as_dict: Literal[False] = False,
        async_retriever_kwargs: dict[str, Any] | None = ...,
    ) -> pd.DataFrame: ...

    @classmethod
    @overload
    def data_dictionary(
        cls,
        data_type: str,
        as_dict: Literal[True] = True,
        async_retriever_kwargs: dict[str, Any] | None = ...,
    ) -> dict[str, Any]: ...

    @classmethod
    def data_dictionary(
        cls,
        data_type: str,
        as_dict: bool = False,
        async_retriever_kwargs: dict[str, Any] | None = None,
    ) -> pd.DataFrame | dict[str, Any]:
        """Retrieve data dictionaries from the STN Flood Event Data API.

        Parameters
        ----------
        data_type : str
            The data source from STN Flood Event Data API.
            It can be ``instruments``, ``peaks``, ``hwms``, or ``sites``.
        as_dict : bool, default = False
            If True, return the data dictionary as a dictionary.
            Otherwise, it returns as ``pandas.DataFrame``.
        async_retriever_kwargs : dict, optional
            Additional keyword arguments to pass to
            ``async_retriever.retrieve_json()``. The ``url`` and ``request_kwds``
            options are already set.

        Returns
        -------
        pandas.DataFrame or dict
            The retrieved data dictionary as pandas.DataFrame or dict.

        See Also
        --------
        :meth:`~get_all_data` : Retrieves all data for a given data type.
        :meth:`~get_filtered_data` : Retrieves filtered data for a given data type.

        Examples
        --------
        >>> from pygeohydro import STNFloodEventData
        >>> data = STNFloodEventData.data_dictionary(data_type="instruments", as_dict=False)  # doctest: +SKIP
        >>> data.shape[1]  # doctest: +SKIP
        2
        >>> data.columns  # doctest: +SKIP
        Index(['Field', 'Definition'], dtype='object')
        """
        dtype_dict = {
            "instruments": "Instruments.csv",
            "peaks": "FilteredPeaks.csv",
            "hwms": "FilteredHWMs.csv",
            "sites": "sites.csv",
        }

        try:
            endpoint = dtype_dict[data_type]
        except KeyError as ke:
            raise InputValueError(data_type, list(dtype_dict.keys())) from ke

        if async_retriever_kwargs is None:
            async_retriever_kwargs = {}
        else:
            async_retriever_kwargs.pop("url", None)

        resp = ar.retrieve_binary(
            [f"{cls.data_dictionary_url}{endpoint}"], **async_retriever_kwargs
        )
        data = pd.read_csv(BytesIO(resp[0]), encoding="latin-1")

        if "Field" not in data.columns:
            data.iloc[0] = data.columns.tolist()
            data.columns = ["Field", "Definition"]

        data["Definition"] = data["Definition"].str.replace("\r\n", "  ")

        # concatenate definitions corresponding to NaN fields until
        # a non-NaN field is encountered
        data_dict = {"Field": [], "Definition": []}

        for f, d in data[["Field", "Definition"]].itertuples(index=False, name=None):
            if pd.isna(f):
                data_dict["Definition"][-1] += " " + d
            else:
                data_dict["Field"].append(f)
                data_dict["Definition"].append(d)

        if as_dict:
            return data_dict
        return pd.DataFrame(data_dict)

    @classmethod
    @overload
    def get_all_data(
        cls,
        data_type: str,
        as_list: Literal[False] = False,
        crs: CRSTYPE = ...,
        async_retriever_kwargs: dict[str, Any] | None = ...,
    ) -> gpd.GeoDataFrame | pd.DataFrame: ...

    @classmethod
    @overload
    def get_all_data(
        cls,
        data_type: str,
        as_list: Literal[True] = True,
        crs: CRSTYPE = ...,
        async_retriever_kwargs: dict[str, Any] | None = ...,
    ) -> list[dict[str, Any]]: ...

    @classmethod
    def get_all_data(
        cls,
        data_type: str,
        as_list: bool = False,
        crs: CRSTYPE = 4326,
        async_retriever_kwargs: dict[str, Any] | None = None,
    ) -> gpd.GeoDataFrame | pd.DataFrame | list[dict[str, Any]]:
        """Retrieve all data from the STN Flood Event Data API.

        Parameters
        ----------
        data_type : str
            The data source from STN Flood Event Data API.
            It can be ``instruments``, ``peaks``, ``hwms``, or ``sites``.
        as_list : bool, optional
            If True, return the data as a list, defaults to False.
        crs : int, str, or pyproj.CRS, optional
            Desired Coordinate reference system (CRS) of output.
            Only used for GeoDataFrames with ``hwms`` and ``sites`` data types.
        async_retriever_kwargs : dict, optional
            Additional keyword arguments to pass to
            ``async_retriever.retrieve_json()``. The ``url`` and ``request_kwds``
            options are already set.

        Returns
        -------
        geopandas.GeoDataFrame or pandas.DataFrame or list of dict
            The retrieved data as a GeoDataFrame, DataFrame, or a list of dictionaries.

        Raises
        ------
        InputValueError
            If the input data_type is not one of
            ``instruments``, ``peaks``, ``hwms``, or ``sites``

        See Also
        --------
        :meth:`~get_filtered_data` : Retrieves filtered data for a given data type.
        :meth:`~data_dictionary` : Retrieves the data dictionary for a given data type.

        Notes
        -----
        Notice schema differences between the data dictionaries, filtered data
        queries, and all data queries. This is a known issue and is being addressed
        by USGS.

        Examples
        --------
        >>> from pygeohydro.stnfloodevents import STNFloodEventData
        >>> data = STNFloodEventData.get_all_data(data_type="instruments")
        >>> data.shape[1]
        18
        >>> data.columns
        Index(['instrument_id', 'sensor_type_id', 'deployment_type_id',
               'location_description', 'serial_number', 'interval', 'site_id',
               'event_id', 'inst_collection_id', 'housing_type_id', 'sensor_brand_id',
               'vented', 'instrument_status', 'data_files', 'files', 'last_updated',
               'last_updated_by', 'housing_serial_number'],
               dtype='object')
        """
        endpoint_dict = {
            "instruments": "Instruments.json",
            "peaks": "PeakSummaries.json",
            "hwms": "HWMs.json",
            "sites": "Sites.json",
        }

        try:
            endpoint = endpoint_dict[data_type]
        except KeyError as ke:
            raise InputValueError(data_type, list(endpoint_dict.keys())) from ke

        if async_retriever_kwargs is None:
            async_retriever_kwargs = {}
        else:
            _ = async_retriever_kwargs.pop("url", None)

        resp = ar.retrieve_json([f"{cls.service_url}{endpoint}"], **async_retriever_kwargs)
        data = [cls._delist_dict(d) for d in resp[0]]  # pyright: ignore[reportArgumentType]

        if as_list:
            return data

        _xy_cols = {
            "instruments": None,
            "peaks": None,
            "hwms": ("longitude_dd", "latitude_dd"),
            "sites": ("longitude_dd", "latitude_dd"),
        }
        xy_cols = _xy_cols[data_type]
        if xy_cols is None:
            return pd.DataFrame(data)

        x_col, y_col = xy_cols
        return cls._geopandify(data, x_col, y_col, crs, cls.service_crs)

    @classmethod
    @overload
    def get_filtered_data(
        cls,
        data_type: str,
        query_params: dict[str, Any] | None = ...,
        as_list: Literal[False] = False,
        crs: CRSTYPE = ...,
        async_retriever_kwargs: dict[str, Any] | None = ...,
    ) -> gpd.GeoDataFrame | pd.DataFrame: ...

    @classmethod
    @overload
    def get_filtered_data(
        cls,
        data_type: str,
        query_params: dict[str, Any] | None = ...,
        as_list: Literal[True] = True,
        crs: CRSTYPE = ...,
        async_retriever_kwargs: dict[str, Any] | None = ...,
    ) -> list[dict[str, Any]]: ...

    @classmethod
    def get_filtered_data(
        cls,
        data_type: str,
        query_params: dict[str, Any] | None = None,
        as_list: bool = False,
        crs: CRSTYPE = 4326,
        async_retriever_kwargs: dict[str, Any] | None = None,
    ) -> gpd.GeoDataFrame | pd.DataFrame | list[dict[str, Any]]:
        """Retrieve filtered data from the STN Flood Event Data API.

        Parameters
        ----------
        data_type : str
            The data source from STN Flood Event Data API.
            It can be ``instruments``, ``peaks``, ``hwms``, or ``sites``.
        query_params : dict, optional
            RESTFUL API query parameters. For accepted values, see
            the STNFloodEventData class attributes :attr:`~instruments_query_params`,
            :attr:`~peaks_query_params`, :attr:`~hwms_query_params`, and
            :attr:`~sites_query_params` for available values.

            Also, see the API documentation for each data type for more information:
                - `instruments <https://stn.wim.usgs.gov/STNServices/Documentation/Sensor/FilteredSensors>`__
                - `peaks <https://stn.wim.usgs.gov/STNServices/Documentation/PeakSummary/FilteredPeakSummaries>`__
                - `hwms <https://stn.wim.usgs.gov/STNServices/Documentation/HWM/FilteredHWMs>`__
                - `sites <https://stn.wim.usgs.gov/STNServices/Documentation/Site/FilteredSites>`__

        as_list : bool, optional
            If True, return the data as a list, defaults to False.
        crs : int, str, or pyproj.CRS, optional
            Desired Coordinate reference system (CRS) of output.
            Only used for GeoDataFrames outputs.
        async_retriever_kwargs : dict, optional
            Additional keyword arguments to pass to
            ``async_retriever.retrieve_json()``. The ``url`` and ``request_kwds``
            options are already set.

        Returns
        -------
        geopandas.GeoDataFrame or pandas.DataFrame or list of dict
            The retrieved data as a GeoDataFrame, DataFrame, or a
            list of dictionaries.

        Raises
        ------
        InputValueError
            If the input data_type is not one of
            ``instruments``, ``peaks``, ``hwms``, or ``sites``
        InputValueError
            If any of the input query_params are not in accepted
            parameters (See :attr:`~instruments_query_params`,
            :attr:`~peaks_query_params`, :attr:`~hwms_query_params`,
            or :attr:`~sites_query_params`).

        See Also
        --------
        :meth:`~get_all_data` : Retrieves all data for a given data type.
        :meth:`~data_dictionary` : Retrieves the data dictionary for a given data type.

        Notes
        -----
        Notice schema differences between the data dictionaries,
        filtered data queries, and all data queries. This is a known
        issue and is being addressed by USGS.

        Examples
        --------
        >>> from pygeohydro.stnfloodevents import STNFloodEventData
        >>> query_params = {"States": "SC, CA"}
        >>> data = STNFloodEventData.get_filtered_data(data_type="instruments", query_params=query_params)
        >>> data.shape[1]
        34
        >>> data.columns
        Index(['sensorType', 'deploymentType', 'eventName', 'collectionCondition',
            'housingType', 'sensorBrand', 'statusId', 'timeStamp', 'site_no',
            'latitude', 'longitude', 'siteDescription', 'networkNames', 'stateName',
            'countyName', 'siteWaterbody', 'siteHDatum', 'sitePriorityName',
            'siteZone', 'siteHCollectMethod', 'sitePermHousing', 'instrument_id',
            'sensor_type_id', 'deployment_type_id', 'location_description',
            'serial_number', 'housing_serial_number', 'interval', 'site_id',
            'vented', 'instrument_status', 'data_files', 'files', 'geometry'],
            dtype='object')
        """
        endpoint_dict = {
            "instruments": "Instruments/FilteredInstruments.json",
            "peaks": "PeakSummaries/FilteredPeaks.json",
            "hwms": "HWMs/FilteredHWMs.json",
            "sites": "Sites/FilteredSites.json",
        }

        try:
            endpoint = endpoint_dict[data_type]
        except KeyError as ke:
            raise InputValueError(data_type, list(endpoint_dict.keys())) from ke

        allowed_query_param_dict = {
            "instruments": cls.instruments_query_params,
            "peaks": cls.peaks_query_params,
            "hwms": cls.hwms_query_params,
            "sites": cls.sites_query_params,
        }

        allowed_query_params = allowed_query_param_dict[data_type]

        if query_params is None:
            query_params = {}

        if not set(query_params.keys()).issubset(allowed_query_params):
            raise InputValueError("query_param", list(allowed_query_params))

        if async_retriever_kwargs is None:
            async_retriever_kwargs = {}
        else:
            async_retriever_kwargs.pop("url", None)
            async_retriever_kwargs.pop("request_kwds", None)

        resp = ar.retrieve_json(
            [f"{cls.service_url}{endpoint}"],
            request_kwds=[{"params": query_params}],
            **async_retriever_kwargs,
        )
        data = [cls._delist_dict(d) for d in resp[0]]  # pyright: ignore[reportArgumentType]
        if as_list:
            return data

        xy_cols = {
            "instruments": ("longitude", "latitude"),
            "peaks": ("longitude_dd", "latitude_dd"),
            "hwms": ("longitude", "latitude"),
            "sites": ("longitude_dd", "latitude_dd"),
        }
        x_col, y_col = xy_cols[data_type]
        return cls._geopandify(data, x_col, y_col, crs, cls.service_crs)


def stn_flood_event(
    data_type: str, query_params: dict[str, Any] | None = None
) -> gpd.GeoDataFrame | pd.DataFrame:
    """Retrieve data from the STN Flood Event Data API.

    Parameters
    ----------
    data_type : str
        The data source from STN Flood Event Data API.
        It can be ``instruments``, ``peaks``, ``hwms``, or ``sites``.
    query_params : dict, optional
        RESTFUL API query parameters, defaults to ``None`` which returns
        a ``pandas.DataFrame`` of information about the given ``data_type``.
        For accepted values, see the ``STNFloodEventData`` class attributes
        :attr:`~.STNFloodEventData.instruments_query_params`,
        :attr:`~.STNFloodEventData.peaks_query_params`,
        :attr:`~.STNFloodEventData.hwms_query_params`, and
        :attr:`~.STNFloodEventData.sites_query_params` for available values.

        Also, see the API documentation for each data type for more information:

        - `instruments <https://stn.wim.usgs.gov/STNServices/Documentation/Sensor/FilteredSensors>`__
        - `peaks <https://stn.wim.usgs.gov/STNServices/Documentation/PeakSummary/FilteredPeakSummaries>`__
        - `hwms <https://stn.wim.usgs.gov/STNServices/Documentation/HWM/FilteredHWMs>`__
        - `sites <https://stn.wim.usgs.gov/STNServices/Documentation/Site/FilteredSites>`__

    Returns
    -------
    geopandas.GeoDataFrame or pandas.DataFrame
        The retrieved data as a GeoDataFrame or DataFrame
        (if ``query_params`` is not passed).

    Raises
    ------
    InputValueError
        If the input data_type is not one of
        ``instruments``, ``peaks``, ``hwms``, or ``sites``
    InputValueError
        If any of the input query_params are not in accepted
        parameters.

    References
    ----------
    * `USGS Short-Term Network (STN) <https://stn.wim.usgs.gov/STNWeb/#/>`__
    * `Filtered Sensors API Documentation <https://stn.wim.usgs.gov/STNServices/Documentation/Sensor/FilteredSensors>`__
    * `Peak Summary API Documentation <https://stn.wim.usgs.gov/STNServices/Documentation/PeakSummary/FilteredPeakSummaries>`__
    * `Filtered HWM API Documentation <https://stn.wim.usgs.gov/STNServices/Documentation/HWM/FilteredHWMs>`__
    * `Filtered Sites API Documentation <https://stn.wim.usgs.gov/STNServices/Documentation/Site/FilteredSites>`__
    * `USGS Flood Event Viewer: Providing Hurricane and Flood Response Data <https://www.usgs.gov/mission-areas/water-resources/science/usgs-flood-event-viewer-providing-hurricane-and-flood>`__
    * `A USGS guide for finding and interpreting high-water marks <https://www.usgs.gov/media/videos/a-usgs-guide-finding-and-interpreting-high-water-marks>`__
    * `High-Water Marks and Flooding  <https://www.usgs.gov/special-topics/water-science-school/science/high-water-marks-and-flooding>`__
    * `Identifying and preserving high-water mark data <https://doi.org/10.3133/tm3A24>`__

    Notes
    -----
    Notice schema differences between the data dictionaries,
    filtered data queries, and all data queries. This is a known
    issue and is being addressed by USGS.

    Examples
    --------
    >>> query_params = {"States": "SC, CA"}
    >>> data = stn_flood_event("instruments", query_params=query_params)
    >>> data.shape[1]
    34
    >>> data.columns
    Index(['sensorType', 'deploymentType', 'eventName', 'collectionCondition',
        'housingType', 'sensorBrand', 'statusId', 'timeStamp', 'site_no',
        'latitude', 'longitude', 'siteDescription', 'networkNames', 'stateName',
        'countyName', 'siteWaterbody', 'siteHDatum', 'sitePriorityName',
        'siteZone', 'siteHCollectMethod', 'sitePermHousing', 'instrument_id',
        'sensor_type_id', 'deployment_type_id', 'location_description',
        'serial_number', 'housing_serial_number', 'interval', 'site_id',
        'vented', 'instrument_status', 'data_files', 'files', 'geometry'],
        dtype='object')
    """
    if query_params is None:
        return STNFloodEventData.get_all_data(data_type=data_type)
    return STNFloodEventData.get_filtered_data(data_type=data_type, query_params=query_params)



"""Accessing National Flood Hazard Layers (NLD) through web services."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Union

from pygeohydro.exceptions import InputValueError
from pynhd import AGRBase

if TYPE_CHECKING:
    import pyproj

    CRSTYPE = Union[int, str, pyproj.CRS]

__all__ = ["NLD"]


class NLD(AGRBase):
    """Access National Levee Database (NLD) services.

    Notes
    -----
    For more info visit: https://geospatial.sec.usace.army.mil/server/rest/services/NLD2_PUBLIC/FeatureServer

    Parameters
    ----------
    layer : str, optional
        A valid service layer. Valid layers are:

        - ``boreholes``
        - ``crossings``
        - ``levee_stations``
        - ``piezometers``
        - ``pump_stations``
        - ``relief_wells``
        - ``alignment_lines``
        - ``closure_structures``
        - ``cross_sections``
        - ``embankments``
        - ``floodwalls``
        - ``frm_lines``
        - ``pipe_gates``
        - ``toe_drains``
        - ``leveed_areas``
        - ``system_routes``
        - ``pipes``
        - ``channels``

    outfields : str or list, optional
        Target field name(s), default to "*" i.e., all the fields.
    crs : str, int, or pyproj.CRS, optional
        Target spatial reference, default to ``EPSG:4326``.

    Methods
    -------
    bygeom(geom, geo_crs=4326, sql_clause="", distance=None, return_m=False, return_geom=True)
        Get features within a geometry that can be combined with a SQL where clause.
    byids(field, fids, return_m=False, return_geom=True)
        Get features by object IDs.
    bysql(sql_clause, return_m=False, return_geom=True)
        Get features using a valid SQL 92 WHERE clause.

    Examples
    --------
    >>> from pygeohydro import NLD
    >>> nld = NLD("levee_stations")
    >>> levees = nld.bygeom((-105.914551, 37.437388, -105.807434, 37.522392))
    >>> levees.shape
    (1838, 12)
    """

    def __init__(
        self,
        layer: Literal[
            "boreholes",
            "crossings",
            "levee_stations",
            "piezometers",
            "pump_stations",
            "relief_wells",
            "alignment_lines",
            "closure_structures",
            "cross_sections",
            "embankments",
            "floodwalls",
            "frm_lines",
            "pipe_gates",
            "toe_drains",
            "leveed_areas",
            "system_routes",
            "pipes",
            "channels",
        ],
        outfields: str | list[str] = "*",
        crs: CRSTYPE = 4326,
    ):
        self.valid_layers = {
            "boreholes": "0",
            "crossings": "1",
            "levee_stations": "2",
            "piezometers": "3",
            "pump_stations": "4",
            "relief_wells": "5",
            "alignment_lines": "6",
            "closure_structures": "7",
            "cross_sections": "8",
            "embankments": "9",
            "floodwalls": "10",
            "frm_lines": "11",
            "pipe_gates": "12",
            "toe_drains": "13",
            "leveed_areas": "14",
            "system_routes": "15",
            "pipes": "16",
            "channels": "17",
        }
        _layer = self.valid_layers.get(layer)
        if _layer is None:
            raise InputValueError("layer", list(self.valid_layers))
        base_url = (
            "https://geospatial.sec.usace.army.mil/server/rest/services/NLD2_PUBLIC/FeatureServer"
        )
        super().__init__(
            f"{base_url}/{_layer}",
            None,
            outfields,
            crs,
        )



"""Accessing data from the supported databases through their APIs."""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Mapping, Tuple, Union, cast

import cytoolz.curried as tlz
import geopandas as gpd
import numpy as np
import pandas as pd
import pyproj
import rasterio as rio
import xarray as xr

import pygeoutils as geoutils
from pygeohydro import helpers
from pygeohydro.exceptions import (
    InputTypeError,
    InputValueError,
    MissingCRSError,
    ServiceUnavailableError,
)
from pygeohydro.helpers import Stats
from pygeoogc import WMS, ServiceURL
from pygeoogc import utils as ogc_utils

if TYPE_CHECKING:
    from numbers import Number

    from shapely import MultiPolygon, Polygon

    GTYPE = Union[Polygon, MultiPolygon, Tuple[float, float, float, float]]
    CRSTYPE = Union[int, str, pyproj.CRS]

__all__ = [
    "nlcd_bygeom",
    "nlcd_bycoords",
    "cover_statistics",
    "overland_roughness",
    "nlcd_area_percent",
]


class NLCD:
    """Get data from NLCD database (2021).

    Parameters
    ----------
    years : dict, optional
        The years for NLCD layers as a dictionary, defaults to
        ``{'impervious': [2021], 'cover': [2021], 'canopy': [2021], "descriptor": [2021]}``.
        Layers that are not in years are ignored, e.g., ``{'cover': [2016, 2021]}`` returns
        land cover data for 2016 and 2021.
    region : str, optional
        Region in the US that the input geometries are located, defaults to ``L48``.
        Valid values are ``L48`` (for CONUS), ``HI`` (for Hawaii), ``AK`` (for Alaska),
        and ``PR`` (for Puerto Rico). Both lower and upper cases are acceptable.
    crs : str, int, or pyproj.CRS, optional
        The spatial reference system to be used for requesting the data, defaults to
        ``epsg:4326``.
    ssl : bool, optional
        Whether to use SSL for the connection, defaults to ``True``.
    """

    def __init__(
        self,
        years: Mapping[str, int | list[int]] | None = None,
        region: str = "L48",
        crs: CRSTYPE = 4326,
        ssl: bool = True,
    ) -> None:
        default_years = {
            "impervious": [2021],
            "cover": [2021],
            "canopy": [2021],
            "descriptor": [2021],
        }
        years = default_years if years is None else years
        if not isinstance(years, dict):
            raise InputTypeError("years", "dict", str(default_years))
        self.years = tlz.valmap(lambda x: x if isinstance(x, list) else [x], years)
        self.years = cast("dict[str, list[int]]", self.years)
        self.region = region.upper()
        base_url = ServiceURL().wms.mrlc
        self.valid_crs = ogc_utils.valid_wms_crs(base_url)
        self.crs = pyproj.CRS(crs).to_string().lower()
        if self.crs not in self.valid_crs:
            raise InputValueError("crs", self.valid_crs)
        self.layers = self.get_layers()
        self.units = {"impervious": "%", "cover": "classes", "canopy": "%", "descriptor": "classes"}
        self.types = {"impervious": "f4", "cover": "u1", "canopy": "f4", "descriptor": "u1"}
        self.nodata = {"impervious": np.nan, "cover": 127, "canopy": np.nan, "descriptor": 127}

        self.wms = WMS(
            base_url,
            layers=list(self.layers.values()),
            outformat="image/geotiff",
            crs=self.crs,
            validation=False,
            ssl=ssl,
        )

    def get_layers(self) -> dict[str, str]:
        """Get NLCD layers for the provided years dictionary."""
        valid_regions = ("L48", "HI", "PR", "AK")
        if self.region not in valid_regions:
            raise InputValueError("region", valid_regions)

        nlcd_meta = helpers.nlcd_helper()

        names = ["impervious", "cover", "canopy", "descriptor"]
        avail_years = {n: nlcd_meta[f"{n}_years"] for n in names}

        if any(
            yr not in avail_years[lyr] or lyr not in names
            for lyr, yrs in self.years.items()
            for yr in yrs
        ):
            vals = [f"\n{lyr}: {', '.join(str(y) for y in yr)}" for lyr, yr in avail_years.items()]
            raise InputValueError("years", vals)

        def layer_name(lyr: str, yr: int) -> str:
            if lyr == "canopy":
                if self.region == "L48":
                    return f"nlcd_tcc_conus_{yr}_v2021-4"
                return f"NLCD_{yr}_Tree_Canopy_{self.region}"
            if lyr == "cover":
                return f"NLCD_{yr}_Land_Cover_{self.region}"
            if lyr == "impervious":
                return f"NLCD_{yr}_Impervious_{self.region}"
            if self.region in ("HI", "PR"):
                raise InputValueError("region (descriptor)", ("L48", "AK"))  # noqa: TRY003
            service_lyr = (
                "Impervious_Descriptor" if self.region == "AK" else "Impervious_descriptor"
            )
            return f"NLCD_{yr}_{service_lyr}_{self.region}"

        return {f"{lyr}_{yr}": layer_name(lyr, yr) for lyr, yrs in self.years.items() for yr in yrs}

    def get_map(
        self,
        geometry: Polygon | MultiPolygon,
        resolution: int,
    ) -> xr.Dataset:
        """Get NLCD response and convert it to ``xarray.DataArray``."""
        r_dict = self.wms.getmap_bybox(geometry.bounds, resolution, self.crs)
        gtiff2xarray = tlz.partial(
            geoutils.gtiff2xarray, geometry=geometry, geo_crs=self.crs, nodata=255
        )
        try:
            _ds = gtiff2xarray(r_dict=r_dict)
        except rio.RasterioIOError as ex:
            raise ServiceUnavailableError(self.wms.url) from ex

        ds = _ds.to_dataset() if isinstance(_ds, xr.DataArray) else _ds
        ds.attrs = _ds.attrs
        for lyr_name, lyr in self.layers.items():
            name = lyr_name.split("_")[0]
            ds = ds.rename({lyr: lyr_name})
            ds[lyr_name] = ds[lyr_name].where(ds[lyr_name] < 255, self.nodata[name])
            ds[lyr_name].attrs["units"] = self.units[name]
            ds[lyr_name] = ds[lyr_name].astype(self.types[name])
            ds[lyr_name].attrs["nodatavals"] = (self.nodata[name],)
            ds[lyr_name] = ds[lyr_name].rio.write_nodata(self.nodata[name])
        return ds

    def __repr__(self) -> str:
        """Print the services properties."""
        return self.wms.__repr__()


def nlcd_bygeom(
    geometry: gpd.GeoSeries | gpd.GeoDataFrame,
    resolution: int = 30,
    years: Mapping[str, int | list[int]] | None = None,
    region: str = "L48",
    crs: CRSTYPE = 4326,
    ssl: bool = True,
) -> dict[int | str, xr.Dataset]:
    """Get data from NLCD database (2019).

    Parameters
    ----------
    geometry : geopandas.GeoDataFrame or geopandas.GeoSeries
        A GeoDataFrame or GeoSeries with the geometry to query. The indices are used
        as keys in the output dictionary.
    resolution : float, optional
        The data resolution in meters. The width and height of the output are computed in pixel
        based on the geometry bounds and the given resolution. The default is 30 m which is the
        native resolution of NLCD data.
    years : dict, optional
        The years for NLCD layers as a dictionary, defaults to
        ``{'impervious': [2019], 'cover': [2019], 'canopy': [2019], "descriptor": [2019]}``.
        Layers that are not in years are ignored, e.g., ``{'cover': [2016, 2019]}`` returns
        land cover data for 2016 and 2019.
    region : str, optional
        Region in the US that the input geometries are located, defaults to ``L48``.
        Valid values are ``L48`` (for CONUS), ``HI`` (for Hawaii), ``AK`` (for Alaska),
        and ``PR`` (for Puerto Rico). Both lower and upper cases are acceptable.
    crs : str, int, or pyproj.CRS, optional
        The spatial reference system to be used for requesting the data, defaults to
        ``epsg:4326``.
    ssl : bool, optional
        Whether to use SSL for the connection, defaults to ``True``.

    Returns
    -------
    dict of xarray.Dataset or xarray.Dataset
        A single or a ``dict`` of NLCD datasets. If dict, the keys are indices
        of the input ``GeoDataFrame``.
    """
    if resolution < 30:
        warnings.warn(
            "NLCD's resolution is 30 m, so finer resolutions are not recommended.",
            UserWarning,
            stacklevel=2,
        )

    if not isinstance(geometry, (gpd.GeoDataFrame, gpd.GeoSeries)):
        raise InputTypeError("geometry", "GeoDataFrame or GeoSeries")

    if geometry.crs is None:
        raise MissingCRSError
    _geometry = cast("gpd.GeoDataFrame", geometry.to_crs(crs))
    geo_dict = _geometry.geometry.to_dict()

    nlcd_wms = NLCD(years=years, region=region, crs=crs, ssl=ssl)

    return {i: nlcd_wms.get_map(g, resolution) for i, g in geo_dict.items()}


def nlcd_bycoords(
    coords: list[tuple[float, float]],
    years: Mapping[str, int | list[int]] | None = None,
    region: str = "L48",
    ssl: bool = True,
) -> gpd.GeoDataFrame:
    """Get data from NLCD database (2019).

    Parameters
    ----------
    coords : list of tuple
        List of coordinates in the form of (longitude, latitude).
    years : dict, optional
        The years for NLCD layers as a dictionary, defaults to
        ``{'impervious': [2019], 'cover': [2019], 'canopy': [2019], "descriptor": [2019]}``.
        Layers that are not in years are ignored, e.g., ``{'cover': [2016, 2019]}`` returns
        land cover data for 2016 and 2019.
    region : str, optional
        Region in the US that the input geometries are located, defaults to ``L48``.
        Valid values are ``L48`` (for CONUS), ``HI`` (for Hawaii), ``AK`` (for Alaska),
        and ``PR`` (for Puerto Rico). Both lower and upper cases are acceptable.
    ssl : bool, optional
        Whether to use SSL for the connection, defaults to ``True``.

    Returns
    -------
    geopandas.GeoDataFrame
        A GeoDataFrame with the NLCD data and the coordinates.
    """
    nlcd_wms = NLCD(years=years, region=region, crs=3857, ssl=ssl)
    points = gpd.GeoSeries(gpd.points_from_xy(*zip(*coords), crs=4326))
    points_proj = points.to_crs(nlcd_wms.crs)
    geoms = points_proj.buffer(50, cap_style="square")
    ds_list = [nlcd_wms.get_map(g, 30) for g in geoms]

    def get_value(da: xr.DataArray, x: float, y: float) -> Number:
        nodata = da.attrs["nodatavals"][0]
        value = da.fillna(nodata).interp(x=[x], y=[y], method="nearest")
        return da.dtype.type(value)[0, 0]

    values = {
        v: [get_value(ds[v], p.x, p.y) for ds, p in zip(ds_list, points_proj)] for v in ds_list[0]
    }
    points = cast("gpd.GeoDataFrame", points.to_frame("geometry"))
    return gpd.GeoDataFrame(
        pd.merge(points, pd.DataFrame(values), left_index=True, right_index=True)
    )


def overland_roughness(cover_da: xr.DataArray) -> xr.DataArray:
    """Estimate overland roughness from land cover data.

    Parameters
    ----------
    cover_da : xarray.DataArray
        Land cover DataArray from a LULC Dataset from the ``nlcd_bygeom`` function.

    Returns
    -------
    xarray.DataArray
        Overland roughness
    """
    if not isinstance(cover_da, xr.DataArray):
        raise InputTypeError("cover_da", "xarray.DataArray")

    roughness = cover_da.astype(np.float64)
    roughness = roughness.rio.write_nodata(np.nan)
    roughness.name = "roughness"
    roughness.attrs["long_name"] = "overland roughness"
    roughness.attrs["units"] = "-"

    meta = helpers.nlcd_helper()
    get_roughness = np.vectorize(meta["roughness"].get, excluded=["default"])
    return roughness.copy(data=get_roughness(cover_da.astype("uint8").astype(str), np.nan))


def cover_statistics(cover_da: xr.DataArray) -> Stats:
    """Percentages of the categorical NLCD cover data.

    Parameters
    ----------
    cover_da : xarray.DataArray
        Land cover DataArray from a LULC Dataset from the ``nlcd_bygeom`` function.

    Returns
    -------
    Stats
        A named tuple with the percentages of the cover classes and categories.
    """
    if not isinstance(cover_da, xr.DataArray):
        raise InputTypeError("cover_da", "xarray.DataArray")

    nlcd_meta = helpers.nlcd_helper()
    val, freq = np.unique(cover_da, return_counts=True)
    zero_idx = np.argwhere(val == 127)
    val = np.delete(val, zero_idx).astype(str)
    freq = np.delete(freq, zero_idx)
    freq_dict = dict(zip(val.tolist(), freq.tolist()))
    total_count = freq.sum()

    if any(c not in nlcd_meta["classes"] for c in freq_dict):
        raise InputValueError("ds", list(nlcd_meta["classes"]))

    class_percentage = {
        nlcd_meta["classes"][k].split(" -")[0].strip(): v / total_count * 100.0
        for k, v in freq_dict.items()
    }
    category_percentage = {
        k: sum(freq_dict[c] for c in v if c in freq_dict) / total_count * 100.0
        for k, v in nlcd_meta["categories"].items()
        if k != "Background"
    }

    return Stats(class_percentage, category_percentage)


def _area_percent(nlcd: xr.Dataset, year: int) -> dict[str, float]:
    """Calculate the percentage of the area for each land cover class."""
    cover_nodata = nlcd[f"cover_{year}"].rio.nodata
    if np.isnan(cover_nodata):
        msk = ~nlcd[f"cover_{year}"].isnull()
    elif cover_nodata > 0:
        msk = nlcd[f"cover_{year}"] < cover_nodata
    else:
        msk = nlcd[f"cover_{year}"] > cover_nodata
    cell_total = msk.sum()

    msk = nlcd[f"cover_{year}"].isin(range(21, 25))
    urban = msk.sum() / cell_total
    natural = 1 - urban

    impervious = nlcd.where(msk)[f"impervious_{year}"].mean() * urban / 100
    developed = urban - impervious
    natural = natural.compute().item() * 100
    developed = developed.compute().item() * 100
    impervious = impervious.compute().item() * 100
    return {
        "natural": natural,
        "developed": developed,
        "impervious": impervious,
        "urban": developed + impervious,
    }


def nlcd_area_percent(
    geo_df: gpd.GeoSeries | gpd.GeoDataFrame,
    year: int = 2019,
    region: str = "L48",
) -> pd.DataFrame:
    """Compute the area percentages of the natural, developed, and impervious areas.

    Notes
    -----
    This function uses imperviousness and land use/land cover data from NLCD
    to compute the area percentages of the natural, developed, and impervious areas.
    It considers land cover classes of 21 to 24 as urban and the rest as natural.
    Then, uses imperviousness percentage to partition the urban area into developed
    and impervious areas. So, ``urban = developed + impervious`` and always
    ``natural + urban = natural + developed + impervious = 100``.

    Parameters
    ----------
    geo_df : geopandas.GeoDataFrame or geopandas.GeoSeries
        A GeoDataFrame or GeoSeries with the geometry to query. The indices are used
        as keys in the output dictionary.
    year : int, optional
        Year of the NLCD data, defaults to 2019. Available years are 2021, 2019, 2016,
        2013, 2011, 2008, 2006, 2004, and 2001.
    region : str, optional
        Region in the US that the input geometries are located, defaults to ``L48``.
        Valid values are ``L48`` (for CONUS), ``HI`` (for Hawaii), ``AK`` (for Alaska),
        and ``PR`` (for Puerto Rico). Both lower and upper cases are acceptable.

    Returns
    -------
    pandas.DataFrame
        A dataframe with the same index as input ``geo_df`` and columns are the area
        percentages of the natural, developed, impervious, and urban
        (sum of developed and impervious) areas. Sum of urban and natural percentages
        is always 100, as well as the sum of natural, developed, and impervious
        percentages.
    """
    valid_year = (2021, 2019, 2016, 2013, 2011, 2008, 2006, 2004, 2001)
    if year not in valid_year:
        raise InputValueError("year", valid_year)

    if not isinstance(geo_df, (gpd.GeoDataFrame, gpd.GeoSeries)):
        raise InputTypeError("geometry", "GeoDataFrame or GeoSeries")

    if geo_df.crs is None:
        raise MissingCRSError

    geoms = geo_df.to_crs(4326).geometry  # pyright: ignore[reportOptionalMemberAccess]

    wms = NLCD(years={"impervious": year, "cover": year}, region=region, ssl=False)

    return pd.DataFrame.from_dict(
        {i: _area_percent(wms.get_map(g, 30), year) for i, g in geoms.items()},
        orient="index",
    )



"""Accessing data from the supported databases through their APIs."""

from __future__ import annotations

import contextlib
import functools
import importlib.util
import io
import itertools
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, Literal, Tuple, Union, cast
from unittest.mock import patch
from zipfile import ZipFile

import cytoolz.curried as tlz
import geopandas as gpd
import numpy as np
import numpy.typing as npt
import pandas as pd
import pyproj
import rasterio as rio
import xarray as xr
from rasterio.io import MemoryFile
from rioxarray import _io as rxr
from shapely.errors import GEOSException

import async_retriever as ar
import pygeoogc as ogc
import pygeoutils as geoutils
from pygeohydro import helpers
from pygeohydro.exceptions import (
    DependencyError,
    InputTypeError,
    InputValueError,
    MissingColumnError,
    ServiceError,
    ZeroMatchedError,
)
from pygeoogc import RetrySession, ServiceURL
from pygeoutils import EmptyResponseError
from pynhd.core import AGRBase, ScienceBase

if TYPE_CHECKING:
    from shapely import MultiPolygon, Polygon

    GTYPE = Union[Polygon, MultiPolygon, Tuple[float, float, float, float]]
    CRSTYPE = Union[int, str, pyproj.CRS]

__all__ = [
    "get_camels",
    "ssebopeta_bycoords",
    "ssebopeta_bygeom",
    "soil_properties",
    "soil_gnatsgo",
    "soil_properties",
    "soil_soilgrids",
    "EHydro",
]


SG_ATTRS = {
    "bdod": {
        "description": "Bulk density of the fine earth fraction",
        "long_name": "Bulk Density",
        "mapped_units": "cg/cm³",
        "conversion_factor": 100,
        "conventional_units": "kg/dm³",
    },
    "cec": {
        "description": "Cation Exchange Capacity of the soil",
        "long_name": "Cation Exchange Capacity",
        "mapped_units": "mmol(c)/kg",
        "conversion_factor": 10,
        "conventional_units": "cmol(c)/kg",
    },
    "cfvo": {
        "description": "Volumetric fraction of coarse fragments (> 2 mm)",
        "long_name": "Coarse Fragments Vol",
        "mapped_units": "cm3/dm3 (vol‰)",
        "conversion_factor": 10,
        "conventional_units": "cm3/100cm3 (vol%)",
    },
    "clay": {
        "description": "Proportion of clay particles (< 0.002 mm) in the fine earth fraction",
        "long_name": "Clay Content",
        "mapped_units": "g/kg",
        "conversion_factor": 10,
        "conventional_units": "g/100g (%)",
    },
    "nitrogen": {
        "description": "Total nitrogen (N)",
        "long_name": "Nitrogen Content",
        "mapped_units": "cg/kg",
        "conversion_factor": 100,
        "conventional_units": "g/kg",
    },
    "phh2o": {
        "description": "Soil pH",
        "long_name": "Ph In H2O",
        "mapped_units": "pHx10",
        "conversion_factor": 10,
        "conventional_units": "pH",
    },
    "sand": {
        "description": "Proportion of sand particles (> 0.05 mm) in the fine earth fraction",
        "long_name": "Sand Content",
        "mapped_units": "g/kg",
        "conversion_factor": 10,
        "conventional_units": "g/100g (%)",
    },
    "silt": {
        "description": "Proportion of silt particles (≥ 0.002 mm and ≤ 0.05 mm) in the fine earth fraction",
        "long_name": "Silt Content",
        "mapped_units": "g/kg",
        "conversion_factor": 10,
        "conventional_units": "g/100g (%)",
    },
    "soc": {
        "description": "Soil organic carbon content in the fine earth fraction",
        "long_name": "Soil Organic Carbon",
        "mapped_units": "dg/kg",
        "conversion_factor": 10,
        "conventional_units": "g/kg",
    },
    "ocd": {
        "description": "Organic carbon density",
        "long_name": "Organic Carbon Density",
        "mapped_units": "hg/m³",
        "conversion_factor": 10,
        "conventional_units": "kg/m³",
    },
    "ocs": {
        "description": "Organic carbon stocks",
        "long_name": "Organic Carbon Stock",
        "mapped_units": "t/ha",
        "conversion_factor": 10,
        "conventional_units": "kg/m²",
    },
}


def get_camels() -> tuple[gpd.GeoDataFrame, xr.Dataset]:
    """Get streaflow and basin attributes of all 671 stations in CAMELS dataset.

    Notes
    -----
    For more info on CAMELS visit: https://ral.ucar.edu/solutions/products/camels

    Returns
    -------
    tuple of geopandas.GeoDataFrame and xarray.Dataset
        The first is basin attributes as a ``geopandas.GeoDataFrame`` and the second
        is streamflow data and basin attributes as an ``xarray.Dataset``.
    """
    base_url = "/".join(
        (
            "https://thredds.hydroshare.org/thredds/fileServer/hydroshare",
            "resources/658c359b8c83494aac0f58145b1b04e6/data/contents",
        )
    )
    urls = [
        f"{base_url}/camels_attributes_v2.0.feather",
        f"{base_url}/camels_attrs_v2_streamflow_v1p2.nc",
    ]
    resp = ar.retrieve_binary(urls)

    attrs = gpd.read_feather(io.BytesIO(resp[0]))
    qobs = xr.open_dataset(io.BytesIO(resp[1]), engine="h5netcdf")
    qobs["discharge"] = xr.where(qobs["discharge"] < 0, np.nan, qobs["discharge"])

    url = "/".join(
        (
            "https://gist.githubusercontent.com/cheginit",
            "229c83c89eee3801a586bcb3ebb4e825/raw/newman_ids.txt",
        )
    )
    ids = ar.retrieve_text([url])[0].split(",")
    attrs["Newman_2017"] = attrs.index.isin(ids)
    qobs["Newman_2017"] = qobs.station_id.isin(ids)
    desc = ". ".join(
        (
            "Whether station is in Newman et al. (2017) dataset",
            "https://doi.org/10.1175/JHM-D-16-0284.1",
        )
    )
    qobs["Newman_2017"].attrs["description"] = desc
    return attrs, qobs


def ssebopeta_bycoords(
    coords: pd.DataFrame,
    dates: tuple[str, str] | int | list[int],
    crs: CRSTYPE = 4326,
) -> xr.Dataset:
    """Daily actual ET for a dataframe of coords from SSEBop database in mm/day.

    Parameters
    ----------
    coords : pandas.DataFrame
        A dataframe with ``id``, ``x``, ``y`` columns.
    dates : tuple or list, optional
        Start and end dates as a tuple (start, end) or a list of years [2001, 2010, ...].
    crs : str, int, or pyproj.CRS, optional
        The CRS of the input coordinates, defaults to ``epsg:4326``.

    Returns
    -------
    xarray.Dataset
        Daily actual ET in mm/day as a dataset with ``time`` and ``location_id`` dimensions.
        The ``location_id`` dimension is the same as the ``id`` column in the input dataframe.
    """
    if not isinstance(coords, pd.DataFrame):
        raise InputTypeError("coords", "pandas.DataFrame")

    req_cols = ["id", "x", "y"]
    if not set(req_cols).issubset(coords.columns):
        raise MissingColumnError(req_cols)

    _coords = gpd.GeoSeries(
        gpd.points_from_xy(coords["x"], coords["y"], crs=crs), index=coords["id"]
    ).to_crs(4326)
    co_list = list(zip(_coords.x, _coords.y))

    f_list = helpers.get_ssebopeta_urls(dates)
    with patch("socket.has_ipv6", False), RetrySession() as session:

        def _ssebop(url: str) -> list[npt.NDArray[np.float64]]:
            r = session.get(url)
            with ZipFile(io.BytesIO(r.content)) as z, MemoryFile() as memfile:
                memfile.write(z.read(z.filelist[0].filename))
                with memfile.open() as src:
                    return list(src.sample(co_list))

        time, eta = zip(*[(t, _ssebop(url)) for t, url in f_list])
    eta_arr = np.array(eta).reshape(len(time), -1)
    ds = xr.Dataset(
        data_vars={
            "eta": (["time", "location_id"], eta_arr),
            "x": (["location_id"], coords["x"].to_numpy()),
            "y": (["location_id"], coords["y"].to_numpy()),
        },
        coords={
            "time": np.array(time, dtype="datetime64[ns]"),
            "location_id": coords["id"].to_numpy(),
        },
    )
    ds["eta"] = ds["eta"].where(ds["eta"] != 9999, np.nan) * 1e-3
    ds.eta.attrs = {
        "units": "mm/day",
        "long_name": "Actual ET",
        "nodatavals": (np.nan,),
    }
    ds.x.attrs = {"crs": pyproj.CRS(crs).to_string()}
    ds.y.attrs = {"crs": pyproj.CRS(crs).to_string()}
    return ds


def ssebopeta_bygeom(
    geometry: GTYPE,
    dates: tuple[str, str] | int | list[int],
    geo_crs: CRSTYPE = 4326,
) -> xr.DataArray:
    """Get daily actual ET for a region from SSEBop database.

    Notes
    -----
    Since there's still no web service available for subsetting SSEBop, the data first
    needs to be downloaded for the requested period then it is masked by the
    region of interest locally. Therefore, it's not as fast as other functions and
    the bottleneck could be the download speed.

    Parameters
    ----------
    geometry : shapely.Polygon or tuple
        The geometry for downloading clipping the data. For a tuple bbox,
        the order should be (west, south, east, north).
    dates : tuple or list, optional
        Start and end dates as a tuple (start, end) or a list of years [2001, 2010, ...].
    geo_crs : str, int, or pyproj.CRS, optional
        The CRS of the input geometry, defaults to ``epsg:4326``.

    Returns
    -------
    xarray.DataArray
        Daily actual ET within a geometry in mm/day at 1 km resolution
    """
    f_list = helpers.get_ssebopeta_urls(dates)

    try:
        _ = geoutils.geo2polygon(geometry)
    except geoutils.InputTypeError as ex:
        raise InputTypeError("geometry", "(Multi)Polygon or tuple of length 4") from ex

    gtiff2xarray = functools.partial(geoutils.gtiff2xarray, geometry=geometry, geo_crs=geo_crs)
    with patch("socket.has_ipv6", False), RetrySession() as session:

        def _ssebop(t: pd.Timestamp, url: str) -> xr.DataArray | xr.Dataset:
            resp = session.get(url)
            with ZipFile(io.BytesIO(resp.content)) as zfile:
                content = zfile.read(zfile.filelist[0].filename)
                return gtiff2xarray(r_dict={"eta": content}).expand_dims({"time": [t]})

        data = xr.merge(itertools.starmap(_ssebop, f_list))
    eta = data.eta.where(data.eta < data.eta.rio.nodata) * 1e-3
    eta = cast("xr.DataArray", eta)
    eta.attrs.update(
        {
            "units": "mm/day",
            "nodatavals": (np.nan,),
            "crs": 4326,
            "long_name": "Actual ET",
        }
    )
    return eta


def soil_properties(
    properties: list[str] | str = "*", soil_dir: str | Path = "cache"
) -> xr.Dataset:
    """Get soil properties dataset in the United States from ScienceBase.

    Notes
    -----
    This function downloads the source zip files from
    `ScienceBase <https://www.sciencebase.gov/catalog/item/5fd7c19cd34e30b9123cb51f>`__
    , extracts the included ``.tif`` files, and return them as an ``xarray.Dataset``.

    Parameters
    ----------
    properties : list of str or str, optional
        Soil properties to extract, default to "*", i.e., all the properties.
        Available properties are ``awc`` for available water capacity, ``fc`` for
        field capacity, and ``por`` for porosity.
    soil_dir : str or pathlib.Pathlib.Path
        Directory to store zip files or if exists read from them, defaults to
        ``./cache``.
    """
    valid_props = {
        "awc": {"name": "awc", "units": "mm/m", "long_name": "Available Water Capacity"},
        "fc": {"name": "fc", "units": "mm/m", "long_name": "Field Capacity"},
        "por": {"name": "porosity", "units": "mm/m", "long_name": "Porosity"},
    }
    if properties == "*":
        prop = list(valid_props)
    else:
        prop = [properties] if isinstance(properties, str) else properties
        if not all(p in valid_props for p in prop):
            raise InputValueError("properties", list(valid_props))

    soil = ScienceBase.get_file_urls("5fd7c19cd34e30b9123cb51f")
    pat = "|".join(prop)
    _files, urls = zip(*soil[soil.index.str.contains(f"{pat}.*zip")].url.to_dict().items())
    urls = list(urls)

    root = Path(soil_dir)
    root.mkdir(exist_ok=True, parents=True)
    files = [root.joinpath(f) for f in _files]
    files = ogc.streaming_download(urls, fnames=files)
    files = [f for f in files if f is not None]

    def get_tif(file: Path) -> xr.DataArray:
        """Get the .tif file from a zip file."""
        with ZipFile(file) as z:
            try:
                fname = next(f.filename for f in z.filelist if f.filename.endswith(".tif"))
            except StopIteration as ex:
                raise ZeroMatchedError from ex
            ds = rxr.open_rasterio(io.BytesIO(z.read(fname)))
            ds = cast("xr.DataArray", ds)
            ds = ds.squeeze("band", drop=True)
            ds.name = valid_props[file.stem.split("_")[0]]["name"]
            ds.attrs["units"] = valid_props[file.stem.split("_")[0]]["units"]
            ds.attrs["long_name"] = valid_props[file.stem.split("_")[0]]["long_name"]
            return ds

    soil = xr.merge((get_tif(f) for f in files), combine_attrs="drop_conflicts")
    _ = soil.attrs.pop("_FillValue", None)
    _ = soil.attrs.pop("units", None)
    _ = soil.attrs.pop("long_name", None)
    return soil


def _open_tiff(file: Path, name: str) -> xr.DataArray:
    """Open a .tif file."""
    ds = rxr.open_rasterio(file)
    ds = cast("xr.DataArray", ds)
    if "band" in ds:
        ds = ds.squeeze("band", drop=True)
    ds.name = name
    return xr.where(ds != ds.rio.nodata, ds, np.nan).rio.write_nodata(np.nan)


def soil_gnatsgo(layers: list[str] | str, geometry: GTYPE, crs: CRSTYPE = 4326) -> xr.Dataset:
    """Get US soil data from the gNATSGO dataset.

    Notes
    -----
    This function uses Microsoft's Planetary Computer service to get the data.
    The dataset's description and its supported soil properties can be found at:
    https://planetarycomputer.microsoft.com/dataset/gnatsgo-rasters

    Parameters
    ----------
    layers : list of str or str
        Target layer(s). Available layers can be found at the dataset's website
        `here <https://planetarycomputer.microsoft.com/dataset/gnatsgo-rasters>`__.
    geometry : Polygon, MultiPolygon, or tuple of length 4
        Geometry or bounding box of the region of interest.
    crs : int, str, or pyproj.CRS, optional
        The input geometry CRS, defaults to ``epsg:4326``.

    Returns
    -------
    xarray.Dataset
        Requested soil properties.
    """
    try:
        import planetary_computer
        import pystac_client
    except ImportError as ex:
        raise DependencyError("soil_gnatsgo", ["pystac-client", "planetary-computer"]) from ex

    catalog = pystac_client.Client.open(
        "https://planetarycomputer.microsoft.com/api/stac/v1",
        modifier=planetary_computer.sign_inplace,
    )
    bounds = geoutils.geo2polygon(geometry, crs, 4326).bounds
    search = catalog.search(collections=["gnatsgo-rasters"], bbox=bounds)
    lyr_href = tlz.merge_with(
        set, ({n: a.href for n, a in i.assets.items()} for i in search.items())
    )
    layers_ = [layers] if isinstance(layers, str) else list(layers)
    layers_ = [lyr.lower() for lyr in layers_]

    def get_layer(lyr: str) -> xr.DataArray:
        fpaths = ogc.streaming_download(list(lyr_href[lyr]), file_extention="tiff")
        fpaths = [f for f in fpaths if f is not None]
        ds = xr.merge(_open_tiff(f, lyr) for f in fpaths)
        affine = ds.rio.transform(recalc=True)
        with rio.open(fpaths[0]) as src:
            ds_crs = src.crs
        ds = ds.rio.write_transform(affine)
        ds = ds.rio.write_crs(ds_crs)
        ds = ds.rio.write_coordinate_system()
        return ds

    soil = xr.merge((get_layer(lyr) for lyr in layers_), combine_attrs="drop_conflicts")
    poly = geoutils.geo2polygon(geometry, crs, soil.rio.crs)
    soil = geoutils.xarray_geomask(soil, poly, soil.rio.crs)
    _ = soil.attrs.pop("_FillValue", None)
    _ = soil.attrs.pop("units", None)
    _ = soil.attrs.pop("long_name", None)
    return soil


def soil_soilgrids(
    layers: list[str] | str,
    geometry: GTYPE,
    geo_crs: CRSTYPE = 4326,
) -> xr.Dataset:
    """Get soil data from SoilGrids for the area of interest.

    Notes
    -----
    For more information on the SoilGrids dataset, visit
    `ISRIC <https://www.isric.org/explore/soilgrids/faq-soilgrids#What_do_the_filename_codes_mean>`__.

    Parameters
    ----------
    layers : list of str
        SoilGrids layers to get. Available options are:
        ``bdod_*``, ``cec_*``, ``cfvo_*``, ``clay_*``, ``nitrogen_*``, ``ocd_*``,
        ``ocs_*``, ``phh2o_*``, ``sand_*``, ``silt_*``, and ``soc_*`` where ``*``
        is the depth in cm and can be one of ``5``, ``15``, ``30``, ``60``,
        ``100``, or ``200``. For example, ``bdod_5`` is the mean bulk density of
        the fine earth fraction at 0-5 cm depth, and ``bdod_200`` is the mean bulk
        density of the fine earth fraction at 100-200 cm depth.
    geometry : Polygon, MultiPolygon, or tuple of length 4
        Geometry to get DEM within. It can be a polygon or a boundong box
        of form (xmin, ymin, xmax, ymax).
    geo_crs : int, str, of pyproj.CRS, optional
        CRS of the input geometry, defaults to ``epsg:4326``.

    Returns
    -------
    xarray.DataArray
        The request DEM at the specified resolution.
    """
    layers_ = [layers] if isinstance(layers, str) else list(layers)
    valid_depths = {
        "5": "0-5cm",
        "15": "5-15cm",
        "30": "15-30cm",
        "60": "30-60cm",
        "100": "60-100cm",
        "200": "100-200cm",
    }
    valid_layers = [
        f"{layer}_{depth}"
        for layer, depth in itertools.product(SG_ATTRS.keys(), valid_depths.keys())
    ]
    invalid_layers = [layer for layer in layers_ if layer not in valid_layers]
    if invalid_layers:
        raise InputValueError("layers", valid_layers, ", ".join(invalid_layers))
    lyr_names, lyr_depths = zip(*[layer.split("_") for layer in layers_])
    base_url = "https://files.isric.org/soilgrids/latest/data"
    geometry_ = geoutils.geo2polygon(geometry)
    bounds = geometry_.bounds

    def _read_layer(lyr: str, depth: str) -> xr.DataArray:
        """Read a SoilGrids layer."""
        ds = rxr.open_rasterio(f"{base_url}/{lyr}/{lyr}_{depth}_mean.vrt")
        ds = cast("xr.DataArray", ds)
        ds = (
            ds.squeeze(drop=True)
            .rio.clip_box(*bounds, crs=geo_crs)
            .rio.clip([geometry_], crs=geo_crs)
        )
        ds = ds.where(ds != ds.rio.nodata)
        ds = ds.rio.write_nodata(np.nan)

        # Convert mapped units to conventional units
        attributes = SG_ATTRS[lyr]
        ds = ds / attributes["conversion_factor"]

        ds.name = f"{lyr}_{depth.replace('-', '_')}_mean"
        ds.attrs["long_name"] = f"Mean {attributes['long_name']} ({depth})"
        ds.attrs["description"] = attributes["description"]
        ds.attrs["units"] = attributes["conventional_units"]

        return ds

    return xr.merge([_read_layer(lyr, valid_depths[d]) for lyr, d in zip(lyr_names, lyr_depths)])


class EHydro(AGRBase):
    """Access USACE Hydrographic Surveys (eHydro).

    Notes
    -----
    For more info visit: https://navigation.usace.army.mil/Survey/Hydro

    Methods
    -------
    bygeom(geom, geo_crs=4326, sql_clause="", distance=None, return_m=False, return_geom=True)
        Get features within a geometry that can be combined with a SQL where clause.
    byids(field, fids, return_m=False, return_geom=True)
        Get features by object IDs.
    bysql(sql_clause, return_m=False, return_geom=True)
        Get features using a valid SQL 92 WHERE clause.

    Parameters
    ----------
    data_type : str, optional
        Type of the survey data to retrieve, defaults to ``points``.
        Note that the ``points`` data type gets the best available point
        cloud data, i.e., if ``SurveyPointHD`` is available, it will be
        returned, otherwise ``SurveyPoint`` will be returned.
        Available types are:

        - ``points``: Point clouds
        - ``outlines``: Polygons of survey outlines
        - ``contours``: Depth contours
        - ``bathymetry``: Bathymetry data

        Note that point clouds are not available for all surveys.
    cache_dir : str or pathlib.Path, optional
        Directory to store the downloaded raw data, defaults to ``./ehydro_cache``.
    """

    def __init__(
        self,
        data_type: Literal["points", "outlines", "bathymetry", "contours"] = "points",
        cache_dir: str | Path = "ehydro_cache",
    ):
        super().__init__(ServiceURL().restful.ehydro)
        self.data_type = data_type
        self.cache_dir = Path(cache_dir)
        self.cache_dir.mkdir(exist_ok=True, parents=True)
        layer = {
            "points": "SurveyPoint",
            "outlines": "SurveyJob",
            "bathymetry": "Bathymetry_Vector",
            "contours": "ElevationContour_ALL",
        }
        if self.data_type not in layer:
            raise InputValueError("data_type", list(layer))
        self._layer = layer[self.data_type]
        self._error_msg = "No survey data found within the given bound."
        self._engine = "pyogrio" if importlib.util.find_spec("pyogrio") else "fiona"
        if self._engine == "pyogrio":
            import pyogrio

            pyogrio.set_gdal_config_options(
                {"OGR_GEOMETRY_ACCEPT_UNCLOSED_RING": "YES", "OGR_ORGANIZE_POLYGONS": "SKIP"}
            )
            warnings.filterwarnings("ignore", message=".*Non closed ring detected.*")
            warnings.filterwarnings("ignore", message=".*translated to Simple Geometry.*")
            self._get_layers = pyogrio.list_layers
        else:
            import logging

            import fiona  # pyright: ignore[reportMissingImports]

            root_logger = logging.getLogger("fiona")
            root_logger.setLevel(logging.ERROR)
            self._get_layers = fiona.listlayers

        bins = AGRBase(ServiceURL().restful.ehydro_bins)
        extent = bins.client.client.extent
        if extent is None:
            msg = "Unable to get the extent of eHydro"
            raise ServiceError(msg)
        self._survey_grid = bins.bygeom(extent)

        warnings.filterwarnings("ignore", message=".*3D.*")
        warnings.filterwarnings("ignore", message=".*OGR_GEOMETRY_ACCEPT_UNCLOSED_RING.*")

    def __post_process(self, survey: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        urls = survey.loc[
            ~survey["sourcedatalocation"].str.contains("placeholder"), "sourcedatalocation"
        ].to_list()
        if not urls:
            raise ZeroMatchedError(self._error_msg)
        fnames = [Path(self.cache_dir, Path(u).name) for u in urls]
        fnames = ogc.streaming_download(urls, fnames=fnames)
        fnames = [f for f in fnames if f is not None]
        if not fnames:
            raise ZeroMatchedError(self._error_msg)

        def get_depth(fname: Path) -> gpd.GeoDataFrame:
            with ZipFile(fname, "r") as zip_ref:
                gdb = next(
                    (
                        f"zip://{fname}!{n.filename}"
                        for n in zip_ref.filelist
                        if n.filename.endswith(".gdb/")
                    ),
                    None,
                )
                if gdb is None:
                    raise ZeroMatchedError(self._error_msg)

            layers = self._get_layers(gdb)
            if self._layer not in layers:
                raise ZeroMatchedError(self._error_msg)
            if self._layer == "SurveyPoint" and "SurveyPointHD" in layers:
                self._layer = "SurveyPointHD"

            gdf = None
            if self._engine == "pyogrio":
                with contextlib.suppress(GEOSException):
                    gdf = gpd.read_file(gdb, layer=self._layer, engine="pyogrio", use_arrow=True)
            if gdf is None:
                gdf = gpd.read_file(gdb, layer=self._layer)
            return gdf.to_crs(5070)  # pyright: ignore[reportCallIssue,reportReturnType]

        return gpd.GeoDataFrame(pd.concat((get_depth(f) for f in fnames), ignore_index=True))

    def _getfeatures(
        self, oids: Iterator[tuple[str, ...]], return_m: bool = False, return_geom: bool = True
    ) -> gpd.GeoDataFrame:
        """Send a request for getting data based on object IDs.

        Parameters
        ----------
        return_m : bool
            Whether to activate the Return M (measure) in the request, defaults to False.
        return_geom : bool, optional
            Whether to return the geometry of the feature, defaults to ``True``.

        Returns
        -------
        geopandas.GeoDataFrame
            The requested features as a GeoDataFrame.
        """
        try:
            return self.__post_process(
                geoutils.json2geodf(
                    self.client.get_features(oids, return_m, return_geom), self.client.client.crs
                )
            )
        except EmptyResponseError as ex:
            raise ZeroMatchedError(self._error_msg) from ex

    @property
    def survey_grid(self) -> gpd.GeoDataFrame:
        """Full survey availability on hexagonal grid cells of 35 km resolution."""
        return self._survey_grid



"""Tests for PyGeoHydro package."""

from __future__ import annotations

import io
import shutil

import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
import xarray as xr
from shapely import Polygon

import pygeohydro as gh
import pygeoutils as geoutils
import pynhd as nhd
from pygeohydro import NFHL, NID, NLD, NWIS, WBD, EHydro

DEF_CRS = 4326
ALT_CRS = 3542
SID_NATURAL = "01031500"
SID_URBAN = "11092450"
DATES = ("2005-01-01", "2005-01-31")
DATES_LONG = ("2000-01-01", "2009-12-31")
GEOM = Polygon(
    [
        [-69.77, 45.07],
        [-69.31, 45.07],
        [-69.31, 45.45],
        [-69.77, 45.45],
        [-69.77, 45.07],
    ]
)


def assert_close(a: float, b: float) -> None:
    np.testing.assert_allclose(a, b, rtol=1e-3)


class TestNWIS:
    nwis: NWIS = NWIS()

    def test_qobs_dv(self):
        df = self.nwis.get_streamflow(SID_NATURAL, DATES)
        ds = self.nwis.get_streamflow(SID_NATURAL, DATES, to_xarray=True)
        col = f"USGS-{SID_NATURAL}"
        assert_close(df[col].sum().item(), ds.sel(station_id=col).discharge.sum().item())
        assert df.attrs[col]["huc_cd"] == ds.sel(station_id=col).huc_cd.item()

    def test_qobs_mmd(self):
        df = self.nwis.get_streamflow(SID_NATURAL, DATES, mmd=True)
        assert_close(df[f"USGS-{SID_NATURAL}"].sum().item(), 27.6375)

    def test_cst_tz(self):
        q = self.nwis.get_streamflow(["08075000", "11092450"], DATES)
        assert q.index.tz.tzname(None) == "UTC"

    def test_qobs_iv(self):
        iv = self.nwis.get_streamflow(SID_NATURAL, ("2020-01-01", "2020-01-31"), freq="iv")
        dv = self.nwis.get_streamflow(SID_NATURAL, ("2020-01-01", "2020-01-31"), freq="dv")
        assert_close(abs(iv.mean().item() - dv.mean().item()), 0.0539)

    def test_info(self):
        query = {"sites": ",".join([SID_NATURAL])}
        info = self.nwis.get_info(query, expanded=True, nhd_info=True)
        assert_close(info["nhd_areasqkm"].item(), 773.964)
        assert info.hcdn_2009.item()

    def test_info_box(self):
        query = {"bBox": ",".join(f"{b:.06f}" for b in GEOM.bounds)}
        info_box = self.nwis.get_info(query, nhd_info=True)
        assert info_box.shape[0] == 35
        assert info_box["nhd_areasqkm"].isna().sum() == 31

    def test_param_cd(self):
        codes = self.nwis.get_parameter_codes("%discharge%")
        assert (
            codes.loc[codes.parameter_cd == "00060", "parm_nm"].iloc[0]
            == "Discharge, cubic feet per second"
        )

    def test_fillna(self):
        index = pd.date_range("2000-01-01", "2020-12-31", freq="D")
        q = pd.Series(np.ones(index.size), index=index)
        qf = gh.streamflow_fillna(q)
        assert not qf.name
        q.loc[slice("2000-01-01", "2000-01-05")] = np.nan
        qf = gh.streamflow_fillna(q)
        assert np.all(qf == 1)
        qf = gh.streamflow_fillna(q.to_frame("12345678"))
        assert np.all(qf == 1)
        qf = gh.streamflow_fillna(xr.DataArray(q))
        assert np.all(qf == 1)


class TestETA:
    dates = ("2000-01-01", "2000-01-05")
    years = [2010, 2014, 2015]

    def test_coords(self):
        coords = pd.DataFrame(
            [
                ["s1", -72.77, 40.07],
                ["s2", -70.31, 46.07],
                ["s3", -69.31, 45.45],
                ["s4", -69.77, 45.45],
            ],
            columns=["id", "x", "y"],
        )
        ds = gh.ssebopeta_bycoords(coords, dates=self.dates)
        assert_close(ds.eta.sum().item(), 8.625)
        assert ds.eta.isnull().sum().item() == 5

    def test_geom(self):
        eta_g = gh.ssebopeta_bygeom(GEOM, dates=self.dates)
        assert_close(eta_g.mean().values.item(), 0.577)

    def test_get_ssebopeta_urls(self):
        _ = gh.helpers.get_ssebopeta_urls(self.years[0])
        urls_dates = gh.helpers.get_ssebopeta_urls(DATES_LONG)
        urls_years = gh.helpers.get_ssebopeta_urls(self.years)
        assert len(urls_dates) == 3653
        assert len(urls_years) == 1095


class TestNLCD:
    years = {"cover": [2016]}
    res = 1000

    @staticmethod
    def assertion(cover, expected):
        st = gh.cover_statistics(cover)
        assert_close(st.categories["Forest"], expected)

    def test_geodf(self):
        geom = gpd.GeoSeries([GEOM, GEOM], crs=DEF_CRS)
        lulc = gh.nlcd_bygeom(geom, years=self.years, resolution=self.res, crs=ALT_CRS, ssl=False)
        self.assertion(lulc[0].cover_2016, 73.1459)
        self.assertion(lulc[1].cover_2016, 73.1459)
        assert lulc[0].cover_2016.rio.nodata == 127

    def test_coords(self):
        coords = list(GEOM.exterior.coords)
        lulc = gh.nlcd_bycoords(coords, ssl=False)
        assert lulc.cover_2021.sum() == 211

    def test_consistency(self):
        coords = [(-87.11890, 34.70421), (-88.83390, 40.17190), (-95.68978, 38.23926)]
        lulc_m = gh.nlcd_bycoords(coords, ssl=False)
        lulc_s = gh.nlcd_bycoords(coords[:1], ssl=False)
        assert lulc_m.iloc[0]["cover_2021"] == lulc_s.iloc[0]["cover_2021"] == 24

    def test_roughness(self):
        geom = gpd.GeoSeries([GEOM], crs=DEF_CRS)
        lulc = gh.nlcd_bygeom(geom, years=self.years, resolution=self.res, crs=ALT_CRS, ssl=False)
        roughness = gh.overland_roughness(lulc[0].cover_2016)
        assert_close(roughness.mean().item(), 0.3256)

    def test_area(self):
        geom = gpd.GeoSeries([GEOM], crs=DEF_CRS)
        area = gh.nlcd_area_percent(geom)
        assert_close(area[["urban", "natural"]].sum(axis=1), 100)
        assert_close(area[["natural", "developed", "impervious"]].sum(axis=1), 100)


class TestNID:
    nid = NID()
    ids = ["KY01232", "GA02400", "NE04081", "IL55070", "TN05345"]

    def test_suggestion(self):
        dams, contexts = self.nid.get_suggestions("houston", "city")
        assert dams.empty
        assert contexts["suggestion"].to_list() == ["Houston", "Houston Lake"]

    def test_filter(self):
        query_list = [
            {"drainageArea": ["[200 500]"]},
            {"nidId": ["CA01222"]},
        ]
        dam_dfs = self.nid.get_byfilter(query_list)
        assert dam_dfs[0].loc[dam_dfs[0].name == "Prairie Portage"].id.item() == "496613"

    def test_id(self):
        dams = self.nid.inventory_byid(self.ids)
        assert_close(dams.damHeight.max(), 39)

    def test_geom(self):
        dams_geo = self.nid.get_bygeom(GEOM, DEF_CRS)
        bbox = geoutils.geometry_reproject(GEOM.bounds, DEF_CRS, ALT_CRS)
        dams_box = self.nid.get_bygeom(bbox, ALT_CRS)
        name = "Pingree Pond"
        assert (dams_geo.name == name).any()
        assert (dams_box.name == "Pingree Pond").any()

    def test_nation(self):
        assert self.nid.df.shape == (92392, 83)
        assert self.nid.gdf.shape == (92184, 96)


class TestWaterQuality:
    wq: gh.WaterQuality = gh.WaterQuality()

    def test_bbox(self):
        stations = self.wq.station_bybbox(
            (-92.8, 44.2, -88.9, 46.0), {"characteristicName": "Caffeine"}
        )
        assert stations.shape[0] == 83

    def test_distance(self):
        stations = self.wq.station_bydistance(-92.8, 44.2, 30, {"characteristicName": "Caffeine"})
        assert stations.shape[0] == 44

    def test_data(self):
        stations = [
            "USGS-435221093001901",
            "MN040-443119093050101",
            "MN040-443602092510501",
            "MN040-443656092474901",
            "MN048-442839093085901",
            "MN048-442849093085401",
            "MN048-443122093050101",
            "MN048-443128092593201",
            "MN048-443129092592701",
            "MN048-443140093042801",
            "MN048-443141093042601",
        ]
        caff = self.wq.data_bystation(stations, {"characteristicName": "Caffeine"})
        assert caff.shape[0] == 12


def test_wbd():
    wbd = WBD("huc4")
    hudson = wbd.byids("huc4", ["0202", "0203"])
    assert len(",".join(hudson.states).split(",")) == 8


def test_states_lookup():
    codes = gh.helpers.states_lookup_table()
    ca = codes["06"].counties
    la_cd = ca[ca.str.contains("Los")].index[0]
    assert la_cd == "037"


@pytest.mark.xfail(reason="Hydroshare is unstable.")
def test_camels():
    attrs, qobs = gh.get_camels()
    assert attrs.shape[0] == qobs.station_id.shape[0] == 671


def test_interactive_map():
    nwis_kwds = {
        "hasDataTypeCd": "dv",
        "outputDataTypeCd": "dv",
        "parameterCd": "00060",
    }
    m = gh.interactive_map((-69.77, 45.07, -69.31, 45.45), nwis_kwds=nwis_kwds)
    assert len(m.to_dict()["children"]) == 4


def test_plot():
    nwis = NWIS()
    qobs = nwis.get_streamflow([SID_NATURAL, SID_URBAN], DATES_LONG)
    gh.plot.signatures(qobs, precipitation=qobs[f"USGS-{SID_NATURAL}"], output="data/gh.plot.png")
    gh.plot.signatures(qobs[f"USGS-{SID_NATURAL}"], precipitation=qobs[f"USGS-{SID_NATURAL}"])
    _, _, levels = gh.plot.cover_legends()
    shutil.rmtree("data")

    assert levels[-1] == 100


def test_nwis_errors():
    err = gh.helpers.nwis_errors()
    assert err.shape[0] == 7


@pytest.mark.parametrize(
    ("key", "expected"),
    [
        (None, 56),
        (["TX", "ca"], 2),
        ("contiguous", 48),
        ("continental", 49),
        ("commonwealths", 4),
        ("territories", 5),
    ],
)
def test_us_states(key, expected):
    states = gh.helpers.get_us_states(key)
    assert states.shape[0] == expected


def test_full_huc():
    hu16 = gh.huc_wb_full(16)
    assert hu16.shape[0] == 7202


def test_irrigation():
    irr = gh.irrigation_withdrawals()
    assert_close(irr.TW.mean(), 419996.4992)


def test_soil():
    soil = gh.soil_properties("por")
    assert soil.sizes["x"] == 266301


@pytest.mark.xfail(reason="NLD is unstable.")
def test_nld():
    nld = NLD("levee_stations")
    levees = nld.bygeom((-105.914551, 37.437388, -105.807434, 37.522392))
    assert levees.shape == (1838, 12)


def test_gnatsgo():
    layers = ["Tk0_100a", "Soc20_50"]
    geometry = (-95.624515, 30.121598, -95.448253, 30.264074)
    soil = gh.soil_gnatsgo(layers, geometry, 4326)
    assert_close(soil.tk0_100a.mean().item(), 89.848)


def test_soilgrid():
    layers = "bdod_5"
    geometry = (-95.624515, 30.121598, -95.448253, 30.264074)
    soil = gh.soil_soilgrids(layers, geometry, 4326)
    assert_close(soil.bdod_0_5cm_mean.mean().item(), 1.4459)


def test_sensorthings():
    sensor = gh.SensorThings()
    # cond = " and ".join(
    #     ("properties/monitoringLocationType eq 'Stream'", "properties/stateFIPS eq 'US:04'")
    # )
    # odata = sensor.odata_helper(conditionals=cond)
    # df = sensor.query_byodata(odata)
    # assert df.shape[0] == 72

    df = sensor.sensor_info("USGS-09380000")
    assert df["description"].iloc[0] == "Stream"

    df = sensor.sensor_property("Datastreams", "USGS-09380000")
    assert df["observationType"].unique()[0] == "Instantaneous"


def test_show_versions():
    f = io.StringIO()
    gh.show_versions(file=f)
    assert "SYS INFO" in f.getvalue()


def test_ehydro():
    bound = (-122.53, 45.57, -122.52, 45.59)
    ehydro = EHydro("bathymetry")
    bathy = ehydro.bygeom(bound)
    assert_close(bathy["depthMean"].mean(), 25.44946)
    assert ehydro.survey_grid.shape[0] == 1125


class TestNFHL:
    """Test the Natinoal Flood Hazard Layer (NFHL) class."""

    @pytest.mark.parametrize(
        ("service", "layer", "expected_url", "expected_layer"),
        [
            (
                "NFHL",
                "cross-sections",
                "https://hazards.fema.gov/arcgis/rest/services/public/NFHL/MapServer",
                "Cross-Sections (14)",
            ),
            (
                "Prelim_CSLF",
                "floodway change",
                "https://hazards.fema.gov/arcgis/rest/services/CSLF/Prelim_CSLF/MapServer",
                "Floodway Change (2)",
            ),
            (
                "Draft_CSLF",
                "special flood hazard area change",
                "https://hazards.fema.gov/arcgis/rest/services/CSLF/Draft_CSLF/MapServer",
                "Special Flood Hazard Area Change (3)",
            ),
            (
                "Prelim_NFHL",
                "preliminary water lines",
                "https://hazards.fema.gov/arcgis/rest/services/PrelimPending/Prelim_NFHL/MapServer",
                "Preliminary Water Lines (20)",
            ),
            (
                "Pending_NFHL",
                "pending high water marks",
                "https://hazards.fema.gov/arcgis/rest/services/PrelimPending/Pending_NFHL/MapServer",
                "Pending High Water Marks (12)",
            ),
            (
                "Draft_NFHL",
                "draft transect baselines",
                "https://hazards.fema.gov/arcgis/rest/services/AFHI/Draft_FIRM_DB/MapServer",
                "Draft Transect Baselines (13)",
            ),
        ],
    )
    def test_nfhl(self, service, layer, expected_url, expected_layer):
        """Test the NFHL class."""
        nfhl = NFHL(service, layer)
        assert nfhl.service_info.url == expected_url
        assert nfhl.service_info.layer == expected_layer

    def test_nfhl_fail_layer(self):
        """Test the layer argument failures in NFHL init."""
        with pytest.raises(nhd.exceptions.InputValueError):
            NFHL("NFHL", "cross_sections")

    def test_nfhl_fail_service(self):
        """Test the service argument failures in NFHL init."""
        with pytest.raises(gh.exceptions.InputValueError):
            NFHL("NTHL", "cross-sections")

    @pytest.mark.parametrize(
        ("service", "layer", "geom", "expected_gdf_len", "expected_schema"),
        [
            (
                "NFHL",
                "cross-sections",
                (-73.42, 43.48, -72.5, 43.52),
                44,
                [
                    "geometry",
                    "OBJECTID",
                    "DFIRM_ID",
                    "VERSION_ID",
                    "XS_LN_ID",
                    "WTR_NM",
                    "STREAM_STN",
                    "START_ID",
                    "XS_LTR",
                    "XS_LN_TYP",
                    "WSEL_REG",
                    "STRMBED_EL",
                    "LEN_UNIT",
                    "V_DATUM",
                    "PROFXS_TXT",
                    "MODEL_ID",
                    "SEQ",
                    "SOURCE_CIT",
                    "SHAPE.STLength()",
                    "GFID",
                    "GlobalID",
                ],
            ),
        ],
    )
    def test_nfhl_getgeom(self, service, layer, geom, expected_gdf_len, expected_schema):
        """Test the NFHL bygeom method."""
        nfhl = NFHL(service, layer)
        gdf_xs = nfhl.bygeom(geom, geo_crs=4269)
        assert isinstance(gdf_xs, gpd.GeoDataFrame)
        assert len(gdf_xs) >= expected_gdf_len
        assert set(gdf_xs.columns) == set(expected_schema)



"""Plot hydrological signatures.

Plots include daily, monthly and annual hydrograph as well as regime
curve (monthly mean) and flow duration curve.
"""

# pyright: reportGeneralTypeIssues=false
from __future__ import annotations

import contextlib
from dataclasses import dataclass, fields
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar, Union

import folium
import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.colors import BoundaryNorm, ListedColormap

import hydrosignatures as hs
import pygeoutils as geoutils
from pygeohydro import helpers
from pygeohydro.exceptions import InputTypeError
from pygeohydro.nwis import NWIS

if TYPE_CHECKING:
    import pyproj

    DF = TypeVar("DF", pd.DataFrame, pd.Series)
    CRSTYPE = Union[int, str, pyproj.CRS]

__all__ = ["signatures", "prepare_plot_data"]


@dataclass(frozen=True)
class PlotDataType:
    """Data structure for plotting hydrologic signatures."""

    daily: pd.DataFrame
    mean_monthly: pd.DataFrame
    ranked: pd.DataFrame
    titles: dict[str, str]
    units: dict[str, str]

    @classmethod
    def fields(cls) -> tuple[str, ...]:
        """Return the field names of the dataclass."""
        return tuple(field.name for field in fields(cls))


def prepare_plot_data(daily: pd.DataFrame | pd.Series) -> PlotDataType:
    """Generate a structured data for plotting hydrologic signatures.

    Parameters
    ----------
    daily : pandas.Series or pandas.DataFrame
        The data to be processed

    Returns
    -------
    PlotDataType
        Containing ``daily, ``mean_monthly``, ``ranked``, ``titles``,
        and ``units`` fields.
    """
    if isinstance(daily, pd.Series):
        daily = daily.to_frame()
    mean_month = hs.mean_monthly(daily, True)
    ranked = hs.exceedance(daily)

    _titles = [
        "Total Hydrograph (daily)",
        "Regime Curve (monthly mean)",
        "Flow Duration Curve",
    ]
    _units = [
        "mm/day",
        "mm/month",
        "mm/day",
    ]
    _fields = PlotDataType.fields()
    titles = dict(zip(_fields[:-1], _titles))
    units = dict(zip(_fields[:-1], _units))
    return PlotDataType(daily, mean_month, ranked, titles, units)


def _prepare_plot_data(
    daily: pd.DataFrame | pd.Series,
    precipitation: pd.DataFrame | pd.Series | None = None,
) -> tuple[PlotDataType, PlotDataType | None]:
    if not isinstance(daily, (pd.DataFrame, pd.Series)):
        raise InputTypeError("daily", "pandas.DataFrame or pandas.Series")

    discharge = prepare_plot_data(daily)
    if precipitation is not None:
        if isinstance(precipitation, pd.DataFrame) and precipitation.shape[1] == 1:
            prcp = prepare_plot_data(precipitation.squeeze())  # pyright: ignore[reportArgumentType]
        elif isinstance(precipitation, pd.Series):
            prcp = prepare_plot_data(precipitation)
        else:
            raise InputTypeError(
                "precipitation", "pandas.Series or pandas.DataFrame with one column"
            )
    else:
        prcp = None

    return discharge, prcp


def signatures(
    discharge: pd.DataFrame | pd.Series,
    precipitation: pd.Series | None = None,
    title: str | None = None,
    figsize: tuple[int, int] | None = None,
    output: str | Path | None = None,
    close: bool = False,
) -> None:
    """Plot hydrological signatures w/ or w/o precipitation.

    Plots includes daily hydrograph, regime curve (mean monthly) and
    flow duration curve. The input discharges are converted from cms
    to mm/day based on the watershed area, if provided.

    Parameters
    ----------
    discharge : pd.DataFrame or pd.Series
        The streamflows in mm/day. The column names are used as labels
        on the plot and the column values should be daily streamflow.
    precipitation : pd.Series, optional
        Daily precipitation time series in mm/day. If given, the data is
        plotted on the second x-axis at the top.
    title : str, optional
        The plot supertitle.
    figsize : tuple, optional
        The figure size in inches, defaults to (9, 5).
    output : str, optional
        Path to save the plot as png, defaults to ``None`` which means
        the plot is not saved to a file.
    close : bool, optional
        Whether to close the figure.
    """
    qdaily, prcp = _prepare_plot_data(discharge, precipitation)

    daily = qdaily.daily
    figsize = (9, 5) if figsize is None else figsize
    fig = plt.figure(constrained_layout=True, figsize=figsize, facecolor="w")
    gs = fig.add_gridspec(2, 3)

    ax = fig.add_subplot(gs[0, :])
    for c, q in daily.items():
        ax.plot(q, label=c)
    lines, labels = ax.get_legend_handles_labels()
    ax.set_ylabel("$Q$ (mm/day)")
    ax.text(0.02, 0.9, "(a)", transform=ax.transAxes, ha="left", va="center", fontweight="bold")

    if prcp is not None:
        _prcp = prcp.daily.squeeze()
        _prcp = _prcp.loc[daily.index[0] : daily.index[-1]]
        label = "$P$ (mm/day)"
        ax_p = ax.twinx()
        ax_p.grid(False)
        if _prcp.shape[0] > 1000:
            ax_p.plot(
                _prcp,
                alpha=0.7,
                color="g",
                label=label,
            )
        else:
            ax_p.bar(
                _prcp.index,
                _prcp.to_numpy().ravel(),
                alpha=0.7,
                width=1,
                color="g",
                align="edge",
                label=label,
            )
        ax_p.set_ylim(_prcp.max() * 2.5, 0)
        ax_p.set_ylabel(label)
        ax_p.set_xmargin(0)
        lines_p, labels_p = ax_p.get_legend_handles_labels()
        lines.extend(lines_p)
        labels.extend(labels_p)

    ax.set_xmargin(0)
    ax.set_xlabel("")
    ax.legend(
        lines,
        labels,
        bbox_to_anchor=(0.5, -0.2),
        loc="upper center",
        ncol=len(lines),
    )

    ax = fig.add_subplot(gs[1, :-1])
    ax.plot(qdaily.mean_monthly)
    ax.set_xmargin(0)
    ax.set_ylabel("$Q$ (mm/month)")
    ax.text(0.02, 0.9, "(b)", transform=ax.transAxes, ha="left", va="center", fontweight="bold")

    ax = fig.add_subplot(gs[1, 2])
    for col in daily:
        dc = qdaily.ranked[[col, f"{col}_rank"]]
        ax.plot(dc[f"{col}_rank"], dc[col], label=col)

    ax.set_yscale("log")
    ax.set_xlim(0, 100)
    ax.set_xlabel("% Exceedance")
    ax.set_ylabel(rf"$\log(Q)$ ({qdaily.units['ranked']})")
    ax.text(0.02, 0.9, "(c)", transform=ax.transAxes, ha="left", va="center", fontweight="bold")

    if title:
        fig.suptitle(title)

    if output is not None:
        Path(output).parent.mkdir(exist_ok=True, parents=True)
        fig.savefig(output, dpi=300)  # pyright: ignore[reportGeneralTypeIssues]

    if close:
        plt.close(fig)


def descriptor_legends() -> tuple[ListedColormap, BoundaryNorm, list[int]]:
    """Colormap (cmap) and their respective values (norm) for land cover data legends."""
    nlcd_meta = helpers.nlcd_helper()
    bounds = [int(v) for v in nlcd_meta["descriptors"]]
    with contextlib.suppress(ValueError):
        bounds.remove(127)

    cmap = ListedColormap(list(nlcd_meta["colors"].values())[: len(bounds)])
    norm = BoundaryNorm(bounds, cmap.N)
    levels = [*bounds, 30]
    return cmap, norm, levels


def cover_legends() -> tuple[ListedColormap, BoundaryNorm, list[int]]:
    """Colormap (cmap) and their respective values (norm) for land cover data legends."""
    nlcd_meta = helpers.nlcd_helper()
    bounds = list(nlcd_meta["colors"])
    with contextlib.suppress(ValueError):
        bounds.remove(127)

    cmap = ListedColormap(list(nlcd_meta["colors"].values()))
    norm = BoundaryNorm(bounds, cmap.N)
    levels = [*bounds, 100]
    return cmap, norm, levels


def interactive_map(
    bbox: tuple[float, float, float, float],
    crs: CRSTYPE = 4326,
    nwis_kwds: dict[str, Any] | None = None,
) -> folium.Map:
    """Generate an interactive map including all USGS stations within a bounding box.

    Parameters
    ----------
    bbox : tuple
        List of corners in this order (west, south, east, north)
    crs : str, int, or pyproj.CRS, optional
        CRS of the input bounding box, defaults to EPSG:4326.
    nwis_kwds : dict, optional
        Additional keywords to include in the NWIS request as a dictionary like so:
        ``{"hasDataTypeCd": "dv,iv", "outputDataTypeCd": "dv,iv", "parameterCd": "06000"}``.
        Default to None.

    Returns
    -------
    folium.Map
        Interactive map within a bounding box.

    Examples
    --------
    >>> import pygeohydro as gh
    >>> nwis_kwds = {"hasDataTypeCd": "dv,iv", "outputDataTypeCd": "dv,iv"}
    >>> m = gh.interactive_map((-69.77, 45.07, -69.31, 45.45), nwis_kwds=nwis_kwds)
    >>> n_stations = len(m.to_dict()["children"]) - 1
    >>> n_stations
    10
    """
    nwis = NWIS()
    bbox = geoutils.geometry_reproject(bbox, crs, 4326)
    query = {"bBox": ",".join(f"{b:.06f}" for b in bbox)}
    if isinstance(nwis_kwds, dict):
        query.update(nwis_kwds)

    sites = nwis.get_info(query, expanded=True, nhd_info=True)

    sites["coords"] = list(sites[["dec_long_va", "dec_lat_va"]].itertuples(name=None, index=False))
    sites["altitude"] = (
        sites["alt_va"].astype("str") + " ft above " + sites["alt_datum_cd"].astype("str")
    )

    sites["drain_area_va"] = sites["drain_area_va"].astype("str") + " sqmi"
    sites["contrib_drain_area_va"] = sites["contrib_drain_area_va"].astype("str") + " sqmi"
    sites["nhd_areasqkm"] = sites["nhd_areasqkm"].astype("str") + " sqkm"
    for c in ("drain_area_va", "contrib_drain_area_va", "nhd_areasqkm"):
        sites.loc[sites[c].str.contains("nan"), c] = "N/A"

    cols_old = [
        "site_no",
        "station_nm",
        "coords",
        "altitude",
        "huc_cd",
        "drain_area_va",
        "contrib_drain_area_va",
        "nhd_areasqkm",
        "hcdn_2009",
    ]

    cols_new = [
        "Site No.",
        "Station Name",
        "Coordinate",
        "Altitude",
        "HUC8",
        "Drainage Area (NWIS)",
        "Contributing Drainage Area (NWIS)",
        "Drainage Area (GagesII)",
        "HCDN 2009",
    ]

    sites = (
        sites.groupby("site_no")[cols_old[1:]]
        .agg(set)
        .reset_index()
        .rename(columns=dict(zip(cols_old, cols_new)))
    )

    msgs = []
    base_url = "https://waterdata.usgs.gov/nwis/inventory?agency_code=USGS&site_no="
    for row in sites.itertuples(index=False):
        site_no = row[sites.columns.get_loc(cols_new[0])]
        msg = f"<strong>{cols_new[0]}</strong>: {site_no}<br>"
        for col in cols_new[1:]:
            value = ", ".join(str(s) for s in row[sites.columns.get_loc(col)])
            msg += f"<strong>{col}</strong>: {value}<br>"
        msg += f'<a href="{base_url}{site_no}" target="_blank">More on USGS Website</a>'
        msgs.append(msg[:-4])

    sites["msg"] = msgs

    west, south, east, north = bbox
    lon = (west + east) * 0.5
    lat = (south + north) * 0.5

    imap = folium.Map(location=(lat, lon), zoom_start=10)

    for coords, msg in sites[["Coordinate", "msg"]].itertuples(name=None, index=False):
        folium.Marker(
            location=next(iter(coords))[::-1],
            popup=folium.Popup(msg, max_width=250),  # pyright: ignore[reportGeneralTypeIssues]
            icon=folium.Icon(),
        ).add_to(imap)

    return imap
